/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.interf;

import java.lang.reflect.Constructor;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.AOKell;


/**
 * This class is the root class for classes generating proxy Interface
 * instances. This class is a container for definitions shared by all real
 * proxy generators.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
abstract public class Delegator {
    
    /**
     * Generate and instantiate a proxy which implements the same interface as a
     * delegate and that extends the InterfaceImpl class.
     * 
     * @param it        the interface type
     * @param itfOwner  the component associated to the interface
     * @param content   the content to which the call must be delegated
     * @param isFcInternalInterface
     *          true if the generated proxy is an internal interface
     * @return  the proxy
     */
    public Object generate(
            InterfaceType it, Component itfOwner, 
            Object content, boolean isFcInternalInterface ) {
        
        /*
         * Subclasses of Delegator:
         * DelegatorForBoundableInterface and DelegatorForImplementedInterface.
         * 
         * suffix contains BoundableInterface or ImplementedInterface depending
         * on the subclass which is used.
         */
        String classname = Delegator.class.getName();
        String suffix = getClass().getName().substring(classname.length()+3);
        
        String delegateClassname = it.getFcItfSignature();
        String targetClassname = "aokell.generated."+delegateClassname+suffix;
        
        Class gen = null;
        try {
            gen = getClass(targetClassname,delegateClassname);
        }
        catch( ClassNotFoundException cnfe ) {
            throw new RuntimeException(cnfe);
        }
        
        // Instanciate the proxy class
        Object o;
        try {
            // Get the (Component,String,Type,boolean) constructor
            Constructor c =
                gen.getDeclaredConstructor( new Class[]{
                    Component.class, String.class, Type.class, Boolean.TYPE} );
            // Invoke it
             o = c.newInstance( new Object[]{
                    itfOwner,               // component owning the interface
                    it.getFcItfName(),          // interface name
                    it,                         // interface type
                    Boolean.valueOf(isFcInternalInterface)} );
        }
        catch( Throwable t ) {
            t.printStackTrace();
            throw new RuntimeException();
        }
        
        return o;
    }
    
    /**
     * Utility method to retrieve a class which has already been generated.
     */
    protected Class getClass(String targetClassname, String delegateClassname)
    throws ClassNotFoundException {
        
                return AOKell.getLoader().loadClass(targetClassname);
    }
        
}
