/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.tools.interf;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.type.TypeFactory;

import aokell.AOKell;
import aokell.lib.InterfaceImpl;
import aokell.lib.control.attribute.AttributeControllerItf;
import aokell.lib.control.binding.BindingControllerItf;
import aokell.lib.control.component.ComponentItf;
import aokell.lib.control.content.ContentControllerItf;
import aokell.lib.control.lifecycle.LifeCycleControllerItf;
import aokell.lib.control.name.NameControllerItf;
import aokell.lib.control.superc.SuperControllerItf;
import aokell.lib.interf.AOKellGeneratedItf;

/**
 * Starting from an ADL description, this tool determines component interfaces
 * and generates the Java classes which will be used to implement these
 * interfaces. For each interface, two classes may be generated: one for server
 * interfaces of primitive component (so called ImplementedInterface) and one
 * for client interfaces and server interfaces exported by composite components
 * (so called BoundableInterface).
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class InterfaceGenerator {
    
    /** The directory where the generated files are stored. */
    private static String generatedDirname = "generated";
    
    /** Root package for generated files. */
    final private static String packageprefix = "aokell.generated";
    
    private static void usage() {
        System.out.println(
                "java "+InterfaceGenerator.class.getName()+
                " [-d dir] [-system] type[s]");
        System.out.println("Description: ");
        System.out.println("  Generate component interface implementations referenced by the given types");
        System.out.println("  Types may be Fractal ADL ");
        System.out.println("Option:");
        System.out.print("  -d dir : the directory where files will be generated");
        System.out.println("(default generated/)");
        System.out.println("  -system: generate AOKell system component interfaces");
    }
    
    public static void main(String[] args) throws Exception {
        
        if( args.length == 0 ) {
            usage();
            return;
        }
        
        new InterfaceGenerator(args).compile();
    }
    
    public InterfaceGenerator( String[] args ) {
        this.args = args;
    }
    
    private String[] args;
    
    
    public void compile() throws IOException, ClassNotFoundException {        
        
        boolean system = false;
        for (int i = 0; i < args.length; i++) {
            if( args[i].equals("-d") ) {
                if( i == args.length-1 ) {
                    usage();
                    return;
                }
                i++;
                generatedDirname = args[i];
            }
            else if( args[i].equals("-system") ) {
                system = true;
            }
            else {
                try {
                    generateFromJava(args[i]);
                }
                catch( IllegalArgumentException iae ) {
                    /* Not a Java interface. */
                    try {
                        generateFromADL(args[i]);
                    }
                    catch( ADLException ae ) {
                        throw new IllegalArgumentException(
                                args[i]+
                                ": unsupported type. Only ADL and interfaces are supported."
                              );
                    }
                }
            }
        }
        
        /*
         * Interfaces used by AOKell.
         */
        if(system) {
            generateImplementedInterface(GenericFactory.class.getName());
            generateImplementedInterface(TypeFactory.class.getName());
            
            generateImplementedInterface(AttributeControllerItf.class.getName());
            generateImplementedInterface(BindingControllerItf.class.getName());
            generateImplementedInterface(ComponentItf.class.getName());
            generateImplementedInterface(ContentControllerItf.class.getName());
            generateImplementedInterface(LifeCycleControllerItf.class.getName());
            generateImplementedInterface(NameControllerItf.class.getName());
            generateImplementedInterface(SuperControllerItf.class.getName());
        }

        System.out.println("Done.");
    }
    
    /**
     * Generate from the given Java file.
     */
    protected void generateFromJava( String java )
    throws IOException, ClassNotFoundException {
        
        Class cl = AOKell.getLoader().loadClass(java);
        if( ! cl.isInterface() ) {
            throw new IllegalArgumentException(java+" is not an interface");
        }

        generateImplementedInterface(java);
        generateBoundableInterface(java);
    }
    
    /**
     * Generate from the given ADL file.
     */
    protected void generateFromADL( String adl )
    throws ADLException, IOException, ClassNotFoundException {
                
        /*
         * Retrieve interface signatures.
         */
        
        // interfaces is of type Set<String>
        // Set<interface signatures>
        Set interfaces = new HashSet();
        
        Map context = new HashMap();
        context.put("interfaces",interfaces);

        // Backend.fractal must be stored in the same package as ContentResolver 
        Factory f = FactoryFactory.getFactory(
                InterfaceGenerator.class.getPackage().getName()+".Backend");
        f.newComponent(adl,context);
        
        
        /*
         * Generate classes implementing interfaces.
         */
        
        for (Iterator iter = interfaces.iterator(); iter.hasNext();) {
            String interf = (String) iter.next();
            generateImplementedInterface(interf);
            generateBoundableInterface(interf);
        }
    }
    
    protected void generateImplementedInterface( String originterfacename )
    throws IOException, ClassNotFoundException {
        
        String interfacename = packageprefix + "." + originterfacename;
        String classname = interfacename + "ImplementedInterface";
        int lastdot = interfacename.lastIndexOf('.');
        String packagename = interfacename.substring(0,lastdot);
        String shortclassname = classname.substring(lastdot+1);
        
        String dirname = generatedDirname+"/"+packagename.replace('.','/');
        String filename = dirname+"/"+shortclassname+".java";
        new File(dirname).mkdirs();
        System.out.println("Writing into: "+filename);
        
        FileWriter fw = new FileWriter(filename);
        PrintWriter pw = new PrintWriter(fw);
        
        prologue(pw,packagename,classname,originterfacename);        
        
        // Delegate field and setter/getter
        pw.println("    private "+originterfacename+" _delegate;");
        pw.println();
        pw.print("    public void _setDelegate( ");
        pw.print(originterfacename);
        pw.println(" _delegate ) {");
        pw.println("        this._delegate = _delegate;");
        pw.println("    }");
        pw.println();
        pw.print("    public ");
        pw.print(originterfacename);
        pw.println(" _getDelegate() {");
        pw.println("        return _delegate;");
        pw.println("    }");
        pw.println();
        
        Class cl = AOKell.getLoader().loadClass(originterfacename);
        Method[] methods = cl.getMethods();
        for (int i = 0; i < methods.length; i++) {
            Method m = methods[i];
            Class rtype = m.getReturnType();
            
            // Method name
            pw.print("    public ");
            pw.print(getClassName(rtype));
            pw.print(" ");
            pw.print(m.getName());
            
            // Method parameters
            pw.print("( ");
            Class[] ptypes = m.getParameterTypes();
            for (int j = 0; j < ptypes.length; j++) {
                if(j!=0)  pw.print(", ");
                Class ptype = ptypes[j];
                pw.print(getClassName(ptype));
                pw.print(" arg");
                pw.print(j);
            }
            pw.print(" ) ");
            
            // Method exceptions
            Class[] etypes = m.getExceptionTypes();
            if( etypes.length != 0 ) {
                pw.print("throws ");
            }
            for (int j = 0; j < etypes.length; j++) {
                Class etype = etypes[j];
                if(j!=0)  pw.print(", ");
                pw.print(etype.getName());
            }
            
            pw.println(" {");
            
            
            // Call to delegate
            pw.print("        ");
            if( ! rtype.getName().equals("void") ) {
                pw.print("return ");
            }
            pw.print("_delegate.");
            pw.print(m.getName());
            pw.print("(");
            for (int j = 0; j < ptypes.length; j++) {
                if(j!=0)  pw.print(", ");
                pw.print("arg");
                pw.print(j);
            }
            pw.println(");");
            
            pw.println("    }");
        }
        
        epilogue(pw);
        
        pw.close();
        fw.close();
    }
    
    protected void generateBoundableInterface( String originterfacename )
    throws IOException, ClassNotFoundException {
        
        String interfacename = packageprefix + "." + originterfacename;
        String classname = interfacename + "BoundableInterface";
        int lastdot = interfacename.lastIndexOf('.');
        String packagename = interfacename.substring(0,lastdot);
        String shortclassname = classname.substring(lastdot+1);
        
        String dirname = generatedDirname+"/"+packagename.replace('.','/');
        String filename = dirname+"/"+shortclassname+".java";
        new File(dirname).mkdirs();
        System.out.println("Writing into: "+filename);
        
        FileWriter fw = new FileWriter(filename);
        PrintWriter pw = new PrintWriter(fw);
        
        prologue(pw,packagename,classname,originterfacename);        
        
        Class cl = AOKell.getLoader().loadClass(originterfacename);
        Method[] methods = cl.getMethods();
        for (int i = 0; i < methods.length; i++) {
            Method m = methods[i];
            Class rtype = m.getReturnType();
            
            // Method name
            pw.print("    public ");
            pw.print(getClassName(rtype));
            pw.print(" ");
            pw.print(m.getName());
            
            // Method parameters
            pw.print("( ");
            Class[] ptypes = m.getParameterTypes();
            for (int j = 0; j < ptypes.length; j++) {
                if(j!=0)  pw.print(", ");
                Class ptype = ptypes[j];
                pw.print(getClassName(ptype));
                pw.print(" arg");
                pw.print(j);
            }
            pw.print(" ) ");
            
            // Method exceptions
            Class[] etypes = m.getExceptionTypes();
            if( etypes.length != 0 ) {
                pw.print("throws ");
            }
            for (int j = 0; j < etypes.length; j++) {
                Class etype = etypes[j];
                if(j!=0)  pw.print(", ");
                pw.print(etype.getName());
            }
            
            pw.println(" {");
            
            
            // Call to delegate
            pw.print("        ");
            if( ! rtype.getName().equals("void") ) {
                pw.print("return ");
            }
            pw.print("((");
            pw.print(originterfacename);
            pw.print(")");
            pw.print("lookupFc()).");
            pw.print(m.getName());
            pw.print("(");
            for (int j = 0; j < ptypes.length; j++) {
                if(j!=0)  pw.print(", ");
                pw.print("arg");
                pw.print(j);
            }
            pw.println(");");
            
            pw.println("    }");
        }
        
        epilogue(pw);
        
        pw.close();
        fw.close();
    }
    
    protected void prologue(
            PrintWriter pw,
            String packagename, String classname, String interfacename )
    throws IOException {
        
        pw.println("/*");
        pw.print(" * Generated by AOKell InterfaceGenerator on: ");
        pw.println(new Date().toString());
        pw.println(" */");
        pw.println();
        
        int lastdot = classname.lastIndexOf('.');
        String shortclassname = classname.substring(lastdot+1);
        
        pw.println("package "+packagename+";");
        pw.println();
        pw.println("import "+Component.class.getName()+";");
        pw.println("import "+Type.class.getName()+";");
        pw.println("import "+InterfaceImpl.class.getName()+";");
        pw.println("import "+AOKellGeneratedItf.class.getName()+";");
        pw.println();
        pw.println("public class "+shortclassname);
        pw.println("   extends InterfaceImpl ");
        pw.println("   implements AOKellGeneratedItf, "+interfacename+" {");
        pw.println();
        pw.println(
                "    public "+shortclassname+
                "(Component component, String s, Type type, boolean flag) {" );
        pw.println("        super(component, s, type, flag);");
        pw.println("    }");
    }

    protected void epilogue( PrintWriter pw ) throws IOException {
        pw.println("}");
    }
    
    /**
     * Workaround for Class.getName() which returns something like
     * Laokell.lib.control.Controller; for array types.
     */
    protected String getClassName( Class cl ) {
        
        if( cl.isArray() ) {
            return getClassName(cl.getComponentType())+"[]";
        }
        
        return cl.getName();
    }
}
