/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.interf;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.AOKell;


/**
 * This class defines methods shared by all classes which generate at runtime
 * implementations of the Interface interface.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
abstract public class Delegator {
    
    /**
     * The map of already generated proxy classes.
     * Map<String,ClassDefinition>. key=target class name, value=class.
     */
    private Map proxies = new HashMap();
    
    /**
     * Generate and instantiate a proxy which implements the same interface as a
     * delegate and that extends the InterfaceImpl class.
     * 
     * @param it        the interface type
     * @param itfOwner  the component associated to the interface
     * @param content   the content to which the call must be delegated
     * @param isFcInternalInterface
     *          true if the generated proxy is an internal interface
     * @return  the proxy
     */
    public Object generate(
            InterfaceType it, Component itfOwner, 
            Object content, boolean isFcInternalInterface ) {
        
        /*
         * Subclasses of Delegator:
         * DelegatorForBoundableInterface and DelegatorForImplementedInterface.
         * 
         * suffix contains BoundableInterface or ImplementedInterface depending
         * on the subclass which is used.
         */
        String classname = Delegator.class.getName();
        String suffix = getClass().getName().substring(classname.length()+3);
        
        String delegateClassname = it.getFcItfSignature();
        String targetClassname = "aokell.generated."+delegateClassname+suffix;
        
        Class gen = null;
        try {
            gen = getClass(targetClassname,delegateClassname);
        }
        catch( ClassNotFoundException cnfe ) {
            ClassDefinition cd =
                generateDelegator(targetClassname,delegateClassname);
            registerProxy(targetClassname,cd);
            gen = cd.cl;
        }
        
        // Instanciate the proxy class
        Object o;
        try {
            // Get the (Component,String,Type,boolean) constructor
            Constructor c =
                gen.getDeclaredConstructor( new Class[]{
                    Component.class, String.class, Type.class, Boolean.TYPE} );
            // Invoke it
             o = c.newInstance( new Object[]{
                    itfOwner,               // component owning the interface
                    it.getFcItfName(),          // interface name
                    it,                         // interface type
                    Boolean.valueOf(isFcInternalInterface)} );
        }
        catch( Throwable t ) {
            t.printStackTrace();
            String tmpdir = System.getProperty("java.io.tmpdir");
            dumpProxies(tmpdir);
            throw new RuntimeException(
                    "Proxy classes dumped in "+tmpdir+"\n"+
                    "targetClassName: "+targetClassname+
                    " / InterfaceType: "+it.getFcItfName()+
                    " / delegate: "+itfOwner);
        }
        
        return o;
    }
    
    /**
     * Generate a proxy that implements the same interface as a delegate and
     * that extends the InterfaceImpl class.
     */
    abstract protected ClassDefinition generateDelegator(
            String targetClassname, String delegateClassname );

    /**
     * Utility method to retrieve a class which has already been generated.
     */
    protected Class getClass(String targetClassname, String delegateClassname)
    throws ClassNotFoundException {
        
        /*
         * If the USE_CACHE property is set, first try to load the generated
         * class from the classpath (this works only if the class has been
         * generated by a previous run.)
         */
        String cachedir =
            System.getProperty(AOKell.USE_GENERATED_CACHE_PROP_NAME,null);
        if( cachedir != null ) {
            try {
                return AOKell.getLoader().loadClass(targetClassname);
            }
            catch( ClassNotFoundException cnfe) {}
        }

        /*
         * The class has not been found in the cache.
         * Search in the local map or generate.
         */
        ClassDefinition cd;
        if( proxies.containsKey(targetClassname) ) {
            cd = (ClassDefinition) proxies.get(targetClassname);
            return cd.cl;
        }
        
        throw new ClassNotFoundException(targetClassname);
    }
    
    /**
     * Store a generated proxy class into the map of proxy classes.
     */
    protected void registerProxy( String targetClassname, ClassDefinition cd ) {
        
        proxies.put(targetClassname,cd);
        
        String cachedir =
            System.getProperty(AOKell.DUMP_GENERATED_PROP_NAME,null);
        if( cachedir != null ) {
            // If applies, dump the generated class to the cache
            try {
                dumpClassToFile(cachedir,targetClassname,cd.bytecode);
            }
            catch(IOException ioe) {
                throw new RuntimeException(ioe);
            }
        }
    }
    
    /**
     * Utility method to write the bytecode from a class to a file. This method
     * creates the directory structure corresponding to package names.
     * 
     * @param dirname    the root directory name
     * @param classname  the name of the class
     * @param b          the bytecode
     */
    public static void dumpClassToFile(
            String dirname, String classname, byte[] b ) throws IOException {
        
        File dir = new File(dirname);
        if( !dir.exists() || !dir.isDirectory() ) {
            throw new IOException(dir+" must exist and be a directory");
        }
        
        // Create the directory structure
        int lastdot = classname.lastIndexOf('.');
        String filedirname = classname.substring(0,lastdot);
        String filename = classname.substring(lastdot+1);
        File filedir = new File(dir,filedirname.replace('.','/'));
        filedir.mkdirs();
        
        // Write the file
        File classfile = new File(filedir,filename+".class");
        FileOutputStream fos = new FileOutputStream(classfile);
        fos.write(b);
        fos.close();
    }

    /**
     * Dump the generated proxy classes in a given directory.
     * 
     * @param dirname  the directory name where the classes must be generated
     */
    public void dumpProxies( String dirname ) {
        
        for (Iterator iter = proxies.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            String targetClassname = (String) entry.getKey();
            ClassDefinition cd = (ClassDefinition) entry.getValue();
            
            try {
                dumpClassToFile(dirname,targetClassname,cd.bytecode);
            }
            catch( IOException ioe ) {
                throw new RuntimeException(ioe);
            }
        }
    }
    
}
