/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.interf;

import java.lang.reflect.Method;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.AOKell;
import aokell.lib.asm.ASMGeneratorForImplementedInterface;



/**
 * This class generates Interface implementations for components which
 * implement directly the interface. This is typically the case of server
 * interfaces associated to primitive components where the content implements
 * the interface. In such a case, calls to the Interface are delegated to the
 * content.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class DelegatorForImplementedInterface extends Delegator {
    
    /** The singleton instance of itself. */
    private static DelegatorForImplementedInterface singleton =
        new DelegatorForImplementedInterface();
    
    private DelegatorForImplementedInterface() {}
    
    public static DelegatorForImplementedInterface get() {
        return singleton;
    }
    
    /**
     * Generate and instantiate a proxy which implements the same interface as a
     * delegate and that extends the InterfaceImpl class.
     * 
     * @param it        the interface type
     * @param itfOwner  the component associated to the interface
     * @param content   the content to which the call must be delegated
     * @param isFcInternalInterface
     *          true if the generated proxy is an internal interface
     * @return  the proxy
     */
    public Object generate(
            InterfaceType it, Component itfOwner,
            Object content, boolean isFcInternalInterface ) {
        
        // Get the proxy
        Object o = super.generate(it,itfOwner,content,isFcInternalInterface);
        
        try {
            String delegateClassname = it.getFcItfSignature();
            
            // Check for frequently encountered problem
            Class itf = AOKell.getLoader().loadClass(delegateClassname);
            if( ! itf.isAssignableFrom(content.getClass()) ) {
                throw new IllegalArgumentException(
                        "Content class "+content.getClass().getName()+
                        " should implement "+itf.getName()+
                        ". Likely cause: trouble with the component controller."
                        );
            }
            
            // Set the delegate
            Method delegmethod =
                o.getClass().getDeclaredMethod("_setDelegate", new Class[]{itf});
            delegmethod.invoke(o,new Object[]{content});
        }
        catch( Throwable t ) {
            t.printStackTrace();
            String tmpdir = System.getProperty("java.io.tmpdir");
            dumpProxies(tmpdir);
            System.err.println("Proxy classes dumped in "+tmpdir);
            throw new RuntimeException();
        }
        
        return o;
    }

    /**
     * Generate a proxy that implements the same interface as a delegate and
     * that extends the InterfaceImpl class.
     */
    protected ClassDefinition generateDelegator(
            String targetClassname, String delegateClassname ) {
        
        return generator.generateDelegator(targetClassname,delegateClassname);
    }

    private ASMGeneratorForImplementedInterface generator =
        new ASMGeneratorForImplementedInterface();
}
