/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.factory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.lib.control.Controller;
import aokell.lib.control.component.ComponentItf;
import aokell.lib.membrane.BaseType;
import aokell.lib.membrane.MCompositeImpl;
import aokell.lib.membrane.MPrimitiveImpl;
import aokell.lib.membrane.MembraneDef;
import aokell.lib.membrane.Membranes;
import aokell.lib.util.FractalHelper;
import aokell.lib.util.MFractalHelper;

/**
 * The singleton implementation of the membrane factory.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class MembraneFactoryImpl implements MembraneFactoryItf {
    
    private MembraneFactoryImpl() {}
    private static MembraneFactoryItf singleton = new MembraneFactoryImpl();
    public static MembraneFactoryItf get() { return singleton; }
    

    /**
     * Create a new control membrane.
     */
    public Component newFcMembrane(
            Type type, Object controllerDesc, Object contentDesc,
            Object content )
    throws InstantiationException {
        
        /*
         * Control membrane for controllers, i.e. components associated to the
         * mPrimitive and mComposite controller descriptions.
         */
        if( controllerDesc.equals("mPrimitive") ) {
            Component ctrlctrl = new MPrimitiveImpl(type,content);
            
            if( !(content instanceof Controller) ) {
                throw new InstantiationException(
                        "Component-controller class "+
                        content.getClass().getName()+" should implement "+
                        Controller.class.getName());
            }
            
            // Set the reference to the component-controller
            Controller ctrl = (Controller)content;
            ctrl.setFcCompCtrl(ctrlctrl);
            
            return ctrlctrl;
        }
        if( controllerDesc.equals("mComposite") ) {
            return new MCompositeImpl(type,content);
        }
        
        /*
         * Control membrane for regular components.
         * 
         * Instantiate the control membrane which is a composite associated to
         * the given controller description.
         */
        MembraneDef mdef = Membranes.get().getMembraneDef(controllerDesc);
        Component membrane = mdef.newFcStaticMembrane();
        
        /*
         * Retrieve the component controller from the previously instantiated
         * control membrane.
         */
        ComponentItf compctrlitf =
            (ComponentItf) FractalHelper.getFcInterface(membrane,"component");
        
        /*
         * Link the content part and the component controller.
         */
        compctrlitf.setContentPart(content);
        if( content instanceof BaseType ) {
            ((BaseType) content).setFcComponent(compctrlitf);
        }
        
        /*
         * Set the attributes exported by the component controller.
         */
        compctrlitf.setType(type);
        compctrlitf.setControllerDesc(controllerDesc);
        compctrlitf.setContentDesc(contentDesc);
        
        /*
         * Initialize controllers.
         */
        Component[] subs = MFractalHelper.getAllSubComponents(membrane);
        for (int i = 0; i < subs.length; i++) {
            Object subcontent = MFractalHelper.getContent(subs[i]);
            if( subcontent != null ) {
                // subcontent may be null for composites
                if( subcontent instanceof Controller ) {
                    Controller ctrl = (Controller)subcontent;
                    ctrl.initFcCtrl();
                }                
            }
        }
        
        /*
         * Initialize the list of control interfaces implemented by this
         * membrane.
         */
        Object[] itfs = membrane.getFcInterfaces();
        for (int i = 0; i < itfs.length; i++) {
            Interface itf = (Interface) itfs[i];
            
            if( ! itf.getFcItfName().startsWith("//") ) {
                itf = MFractalHelper.followBindingsUpToLastInterface(itf);                
                InterfaceType it = (InterfaceType) itf.getFcItfType();
                Component itfOwner = itf.getFcItfOwner();
                Controller ctrl =
                    (Controller) MFractalHelper.getContent(itfOwner);
                compctrlitf.addFcController(it,ctrl);
            }
        }
        
        /*
         * Re-initialize the component controller.
         * Control interfaces have been updated.
         */
        Interface citf =
            MFractalHelper.followBindingsUpToLastInterface((Interface)compctrlitf);
        Component c = citf.getFcItfOwner();
        Controller compctrl = (Controller) MFractalHelper.getContent(c);
        compctrl.initFcCtrl();
        
        return (Component) compctrl;
    }
    
    /**
     * Return true if the content class associated to the given controller
     * description must implement the BindingController interface when its type
     * defines at least one client interface. 
     */
    public boolean checkFcContentForBC( Object controllerDesc ) {
        if( controllerDesc.equals("mPrimitive") ||
            controllerDesc.equals("mComposite") )
            return false;
        
        return true;
    }

}
