/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.util;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.NameController;

import aokell.lib.control.superc.SuperControllerItf;

/**
 * Helper class for control components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class MFractalHelper {

    /**
     * Returns the {@link BindingController} interface of the given
     * control component.
     */
    public static BindingController getBindingController(Component c) {
        return (BindingController)
            FractalHelper.getFcInterface(c,"//binding-controller");
    }

    /**
     * Returns the {@link ContentController} interface of the given
     * control component.
     */
    public static ContentController getContentController(Component c) {
        return (ContentController)
            FractalHelper.getFcInterface(c,"//content-controller");
    }

    /**
     * Returns the {@link SuperControllerItf} interface of the give
     * control component.
     */
    public static SuperControllerItf getSuperController(Component c) {
        return (SuperControllerItf)
            FractalHelper.getFcInterface(c,"//super-controller");
    }

    /**
     * Returns the {@link NameController} interface of the given 
     * control component.
     */
    public static NameController getNameController(Component c) {
        return (NameController)
            FractalHelper.getFcInterface(c,"//name-controller");
    }

    public static Object getContent(Component c) {
        return FractalHelper.getFcInterface(c,"///content");
    }
    
    /**
     * Follow bindings and return the interface associated to the component
     * located at the end of the binding chain.
     */
    public static Interface followBindingsUpToLastInterface( Interface start ) {
        
        Object next = null;
        Component comp = null;
        
        do {
            comp = start.getFcItfOwner();            
            BindingController bc = getBindingController(comp);
            try {
                next = bc.lookupFc(start.getFcItfName());
                if( next != null ) {
                    start = (Interface) next;
                }
            }
            catch( NoSuchInterfaceException nsie ) {
                next = null;
            }
        }
        while( next != null );
        
        return start;
    }

    /**
     * Return all sub-components of the given component.
     */
    public static Component[] getAllSubComponents( Component c ) {
        List l = new ArrayList();
        addAllSubComponents(c,l);
        return (Component[]) l.toArray( new Component[l.size()] );
    }

    /**
     * Add all sub-components of the given component to the list.
     */
    public static void addAllSubComponents( Component c, List l ) {
        
        if( l == null ) {
            throw new IllegalArgumentException(
                    "The given list (2nd parameter) must be non null");
        }
        
        l.add(c);
        
        try {
            ContentController cc =
                (ContentController) c.getFcInterface("//content-controller");
            Component[] subs = cc.getFcSubComponents();
            for (int i = 0; i < subs.length; i++) {
                addAllSubComponents(subs[i],l);
            }
        }
        catch( NoSuchInterfaceException nsie ) {
            // A non composite component has been reached
        }
        
    }
}
