/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.factory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;

import aokell.lib.control.Controller;
import aokell.lib.control.component.ComponentItf;
import aokell.lib.membrane.BaseType;
import aokell.lib.membrane.MembraneDef;
import aokell.lib.membrane.Membranes;

/**
 * The singleton implementation of the membrane factory.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class MembraneFactoryImpl implements MembraneFactoryItf {
    
    private MembraneFactoryImpl() {}
    private static MembraneFactoryItf singleton = new MembraneFactoryImpl();
    public static MembraneFactoryItf get() { return singleton; }
    

    /**
     * Create a new control membrane.
     */
    public Component newFcMembrane(
            Type type, Object controllerDesc, Object contentDesc,
            Object content )
    throws InstantiationException {
        
        /*
         * Create the component controller.
         */
        ComponentItf compctrl = newFcComponentController(controllerDesc);
        
        /*
         * Link the content part and the component controller.
         */
        compctrl.setContentPart(content);
        if( content instanceof BaseType ) {
            ((BaseType) content).setFcComponent(compctrl);
        }
        
        /*
         * Set the attributes exported by the component controller.
         */
        compctrl.setType(type);
        compctrl.setControllerDesc(controllerDesc);
        compctrl.setContentDesc(contentDesc);
        
        /*
         * Initialize the controllers.
         */
        initFcControllers(controllerDesc,compctrl);
        
        return compctrl;
    }

    
    /**
     * Return true if the content class associated to the given controller
     * description must implement the BindingController interface when its type
     * defines at least one client interface. 
     */
    public boolean checkFcContentForBC( Object controllerDesc ) {
        // Always true for AOKell with pure object-oriented membranes
        return true;
    }

    
    // -----------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------
    
    private ComponentItf newFcComponentController( Object controllerDesc )
    throws InstantiationException {
        
        Membranes mr = Membranes.get();
        MembraneDef mdef = mr.getMembraneDef(controllerDesc);
        ComponentItf compctrl = null;
        
        for (int i = 0; i < mdef.ctrls.length; i++) {
            Class ctrlcl = mdef.ctrls[i].impl;
            
            if( Component.class.isAssignableFrom(ctrlcl) ) {
                Controller ctrl = null;
                try {
                    ctrl = (Controller) ctrlcl.newInstance();
                }
                catch( Exception e ) {
                    throw new InstantiationException(e.getMessage());
                }
                compctrl = (ComponentItf) ctrl;
                ctrl.setFcCompCtrl(compctrl);
                ctrl.initFcCtrl();
                compctrl.addFcController(mdef.ctrls[i].type,ctrl);
            }
        }
        
        return compctrl;
    }
    
    /**
     * Initialize the controllers.
     */
    private void initFcControllers(
            Object controllerDesc, ComponentItf compctrl )
    throws InstantiationException {
        
        Membranes mr = Membranes.get();
        MembraneDef mdef = mr.getMembraneDef(controllerDesc);
        
        for (int i = 0; i < mdef.ctrls.length; i++) {
            Class ctrlcl = mdef.ctrls[i].impl;
            
            // Skip the component controller
            if( ! Component.class.isAssignableFrom(ctrlcl) ) {
                Controller ctrl = null;
                try {
                    ctrl = (Controller) ctrlcl.newInstance();
                }
                catch( Exception e ) {
                    throw new InstantiationException(e.getMessage());
                }
                ctrl.setFcCompCtrl(compctrl);
                ctrl.initFcCtrl();
                compctrl.addFcController(mdef.ctrls[i].type,ctrl);
            }
        }
        
        /*
         * Re-initialize the component controller.
         * Controllers have been added and cached values (list of interfaces,
         * etc.) may need to be recomputed.
         */
        ((Controller)compctrl).initFcCtrl();
    }
    
}
