/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell;


import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.Factory;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;

import aokell.lib.factory.GenericFactoryImpl;
import aokell.lib.membrane.Membranes;
import aokell.lib.type.ComponentTypeImpl;
import aokell.lib.type.InterfaceTypeImpl;


/**
 * The AOKell factory for retrieving the bootstrap component.
 * A Fractal bootstrap component provides the TypeFactory and GenericFactory
 * server interfaces.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class AOKell implements Factory {
    
    /**
     * When non null, this property tells AOKell to dump generated code. This
     * property is used in aokell.lib.asm.
     */
    final public static String DUMP_GENERATED_PROP_NAME = "aokell.dumpgenerated";
    
    /**
     * When non null, this property tells AOKell to use a cache for generated
     * code. This property is used in aokell.lib.asm prior to code generation.
     */
    final public static String USE_GENERATED_CACHE_PROP_NAME = "aokell.usecache";
    
    public Type getFcInstanceType() {
        return TYPE;
    }

    public Object getFcControllerDesc() {
        return CONTROLLER_DESC;
    }

    public Object getFcContentDesc() {
        /*
         * Hard-coding the name of the bootstrap component to avoid a compile
         * dependency from this class to aokell.component which would hinder the
         * modularity of the build process.
         */
        return "aokell.component.BootstrapComponentImpl";
    }

    /** The singleton instance of the bootstrap component. */
    private static Component bootstrap;
    
    /**
     * This method returns the bootstrap component i.e. the component returned
     * when Fractal.getBootstrapComponent() is called.
     * The bootstrap component provides the interfaces:
     * <ul>
     * <li> org.objectweb.fractal.api.type.TypeFactory </li>
     * <li> org.objectweb.fractal.api.factory.GenericFactory </li>
     * </ul>
     */
    public Component newFcInstance() throws InstantiationException {
        
        if( bootstrap != null )
            return bootstrap;
        
        // Register AOKell membranes with the membrane repository
        Membranes.get().register(AOKellMembranes.MEMBRANES);

        // Create the bootstrap component
        GenericFactory gf = GenericFactoryImpl.get();
        bootstrap =
            gf.newFcInstance(
                    getFcInstanceType(),
                    getFcControllerDesc(),
                    getFcContentDesc() );
        
        return bootstrap;
    }

    final public static ComponentType TYPE =
        new ComponentTypeImpl(
                new InterfaceType[] {
                        new InterfaceTypeImpl(
                                "type-factory",
                                TypeFactory.class.getName(),
                                false, false, false),
                        new InterfaceTypeImpl(
                                "generic-factory",
                                GenericFactory.class.getName(),
                                false, false, false)
                });
    
    final public static Object CONTROLLER_DESC = "bootstrap";
    
    /** The loader used by AOKell to load classes. */
    private static ClassLoader loader = ClassLoader.getSystemClassLoader();
    
    /**
     * Set the loader to be used by AOKell for loading classes.
     */
    public static void setLoader( ClassLoader cl ) { loader=cl; }
    
    /**
     * Return the loader used by AOKell for loading classes.
     */
    public static ClassLoader getLoader() { return loader; }
}
