/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.util.Fractal;

/**
 * Default implementation of the {@link Interface} interface.
 * This class is extended by classes which implement component interfaces.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class InterfaceImpl implements Interface, CloneableItf {
    
    private Component fcItfOwner;
    private String fcItfName;
    private Type fcItfType;
    private boolean fcInternalItf;
    

    public InterfaceImpl(Component fcItfOwner, String fcItfName,
            Type fcItfType, boolean fcInternalItf) {
        init(fcItfOwner,fcItfName,fcItfType,fcInternalItf);
    }
    
    public InterfaceImpl() {
    }
    
    public void init(Component fcItfOwner, String fcItfName,
            Type fcItfType, boolean fcInternalItf) {

        setFcItfOwner(fcItfOwner);
        setFcItfName(fcItfName);
        setFcItfType(fcItfType);
        setFcInternalItf(fcInternalItf);
    }
        

    // ------------------------------------------------------------------
    // Implementation of the Interface interface
    // ------------------------------------------------------------------
    
    public Component getFcItfOwner() {
        return fcItfOwner;
    }

    public String getFcItfName() {
        return fcItfName;
    }

    public Type getFcItfType() {
        return fcItfType;
    }

    public boolean isFcInternalItf() {
        return fcInternalItf;
    }

    
    // ------------------------------------------------------------------
    // Implementation specific methods
    // ------------------------------------------------------------------
    
    /**
     * If this interface can play the role of the source interface in a binding,
     * return the target interface currently bound to it (if there is one.)
     * Client interfaces or server interfaces owned by a composite can play the
     * role of the source interface in a binding.
     */
    public Object lookupFc() {
        try {
            return bc.lookupFc(fcItfName);
        }
        catch( NullPointerException npe ) {
            setBC();
            if( bc == null ) {
                throw new RuntimeException(
                        fcItfOwner+" should have a binding controller");
            }
            return lookupFc();
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(
                    "No such interface exception: "+nsie.getMessage());
        }
    }
    
    /** The binding controller associated to the owner of this interface. */
    protected BindingController bc;
    
    protected void setBC() {
        try {
            bc = Fractal.getBindingController(fcItfOwner);
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(
                    "No such interface exception: "+nsie.getMessage());
        }
    }
    
    public void setFcItfOwner( Component fcItfOwner ) {
        this.fcItfOwner = fcItfOwner;
        bc = null;
    }
    
    public void setFcItfName( String fcItfName ) {
        this.fcItfName = fcItfName;
    }
    
    public void setFcItfType( Type fcItfType ) {
        this.fcItfType = fcItfType;
    }
    
    public void setFcInternalItf(boolean fcInternalItf) {
        this.fcInternalItf = fcInternalItf;
    }

    
    // ------------------------------------------------------------------
    // Implementation of the CloneableItf interface
    // ------------------------------------------------------------------
    
    /*
     * clone() is required and could not be replaced by a constructor.
     * We need in ComponentImpl to be able to create instances of sub-classes
     * of InterfaceImpl generated on the fly in aokell.lib.asm. 
     */
    public Object clone() {
        InterfaceImpl clone = null;
        try {
            clone = (InterfaceImpl) super.clone();
        }
        catch(CloneNotSupportedException cnse) {
            throw new RuntimeException(cnse);
        }
        clone.fcItfOwner = fcItfOwner;
        clone.fcItfName = fcItfName;
        clone.fcItfType = fcItfType;
        clone.fcInternalItf = fcInternalItf;
        return clone;
    }
}
