/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.control.attribute;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.AttributeController;

import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.util.FractalHelper;

/**
 * Implementation of the attribute controller.
 * 
 * The role played by this class is very special in sense that when the
 * attribute controller is looked up on a component
 * (e.g. with <code>aComp.getFcInterface("attribute-controller")</code>)
 * this is not this instance that is returned. Indeed the attribute controller
 * is defined in the component type with something like:
 * <code>
 *         ComponentType sType = tf.createFcType(new InterfaceType[] {
 *               tf.createFcItfType("s", "cs.impl.Service", false, false, false),
 *               tf.createFcItfType(
 *                       "attribute-controller",
 *                       "cs.impl.ServiceAttributes",
 *                       false,
 *                       false,
 *                       false)
 *       });
 * </code>
 * Hence the attribute controller is seen as a business server interface.
 * The goal of the current class is to provide an implementation for
 * cloneCtrl(). This method is called when a template is instantiated to
 * transfert the state of the attribute controller to the new instance. Only
 * state information accessible with public getters are transfered. 
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class AttributeControllerImpl
    implements AttributeControllerItf, Controller {
    
    public AttributeControllerImpl() {}
    
    
    // --------------------------------------------------------------
    // AttributeController interface
    // --------------------------------------------------------------
    
    // Empty
    
    // --------------------------------------------------------------
    // Controller implementation
    // --------------------------------------------------------------
    
    private Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        // Indeed nothing
    }
    
    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers.
     * 
     * This method retrieves the state of attributes by invoking all the
     * getters of the current component and by calling the associated setters
     * on the destination component.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {

        Object contentPart = FractalHelper.getContent(compctrl);
        Object dstContentPart = FractalHelper.getContent(dst);

        try {
            initializeAttributes(contentPart,dstContentPart);
            
            // Read-write attributes
            for (Iterator iter = rwAttributes.entrySet().iterator();
                    iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Method getter = (Method) entry.getKey();
                Method setter = (Method) entry.getValue();

                // Invoke the getter on the current component
                Object value = getter.invoke(contentPart,new Object[]{});
                // Invoke the setter on the destination component
                setter.invoke(dstContentPart,new Object[]{value});
            }
            
            // Read-only attributes
            for (Iterator iter = rAttributes.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Method getter = (Method) entry.getKey();
                Field attr = (Field) entry.getValue();
                
                // Invoke the getter on the current component
                Object value = getter.invoke(contentPart,new Object[]{});
                // Set the field on the destination component
                attr.set(dstContentPart,value);
            }
            
            // Write-only attributes
            for (Iterator iter = wAttributes.entrySet().iterator();
                    iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Method setter = (Method) entry.getKey();
                Field attr = (Field) entry.getValue();
                
                // Get the field on the current component
                Object value = attr.get(contentPart);
                // Invoke the setter on the destination component
                setter.invoke(dstContentPart,new Object[]{value});
            }
        }
        catch(Exception e) {
            e.printStackTrace();
            throw new CloneCtrlException(e);
        }
    }
    
    private void initializeAttributes(
            Object contentPart, Object dstContentPart ) {
        
        if( rwAttributes != null )
            return;
        
        rwAttributes = new HashMap();
        rAttributes = new HashMap();
        wAttributes = new HashMap();
        
        // Search for the attribute controller interface
        
        Class cl = contentPart.getClass();
        Class[] itfs = cl.getInterfaces();
        Class cldst = dstContentPart.getClass();
        
        Class acitf = null;
        for (int i = 0; i < itfs.length; i++) {
            if( AttributeController.class.isAssignableFrom(itfs[i]) ) {
                acitf = itfs[i];
                break;
            }
        }
        if( acitf == null ) {
            throw new RuntimeException(
                    "The attribute controller interface should have been found");
        }
        
        // Initialize read-write and read-only attributes
        initializeReadAttributes(acitf,cl,cldst);
        
        // Initialize write-only attributes
        initializeWriteOnlyAttributes(acitf,cl,cldst);
    }
    
    /**
     * @param acitf  the class of the attribute controller Java interface
     * @param cl     the class of the source component
     * @param cldst  the class of the destination component
     */
    private void initializeReadAttributes(Class acitf,Class cl,Class cldst) {
        
        // Get method defined on the attribute controller interface
        Method[] methods = acitf.getMethods();
        
        for (int i = 0; i < methods.length; i++) {
            Method getter = methods[i];
            String getterName = getter.getName(); 
            if( getterName.startsWith("get") ) {
                Class[] params = methods[i].getParameterTypes();
                Class ret = methods[i].getReturnType();
                if( params.length==0 && !ret.equals(Void.TYPE) ) {
                    // type getXXX() where type!=void
                    try {
                        // Lookup the getter on the component class
                        Method realGetter = cl.getMethod(
                                    getterName,new Class[]{});
                        String setterName = "set"+getterName.substring(3);
                        Method setter =
                            cldst.getMethod(setterName,new Class[]{ret});
                        if( setter.getReturnType().equals(Void.TYPE)) {
                            rwAttributes.put(realGetter,setter);
                        }
                    }
                    catch(NoSuchMethodException nsme) {
                        // No associated setter
                        try {
                            Field field = getField(cl,getter);
                            rAttributes.put(getter,field);
                        }
                        catch( NoSuchFieldException nsfe ) {
                            // No associated field.
                            // Can't do anything.
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Get the field accessed from a given getter or getter method. The field
     * name is computed from the getter or setter name.
     * 
     * @param cl        the class
     * @param accessor  the getter or the setter
     * @return          the field
     */
    private static Field getField(Class cl,Method accessor)
    throws NoSuchFieldException {
        
        String accessorName = accessor.getName();
        
        // Skip the "get" or "set" prefix and un-caps the first letter
        StringBuffer sb = new StringBuffer(accessorName.substring(3));
        char first = Character.toLowerCase(sb.charAt(0));
        sb.setCharAt(0,first);
        
        String attrName = sb.toString();
        Field attr = null;
        
        try {
            // getDeclared to get non-public fields
            attr = cl.getDeclaredField(attrName);
            // Suppress Java accessibiliy control check
            attr.setAccessible(true);
        }
        catch( NoSuchFieldException nsfe ) {
           // try to get it as a public inherited field 
           attr = cl.getField(attrName);
           // TODO iterate over inherited classes to get non-public inherited fields
        }
        
        return attr;
    }
    
    private void initializeWriteOnlyAttributes(Class accl,Class cl,Class cldst) {
        
        Method[] methods = accl.getMethods();
        
        for (int i = 0; i < methods.length; i++) {
            Method setter = methods[i];
            String setterName = setter.getName(); 
            if( setterName.startsWith("set") ) {
                Class[] params = methods[i].getParameterTypes();
                Class ret = methods[i].getReturnType();
                if( params.length==1 && ret.equals(Void.TYPE) ) {
                    // void setXXX(type)
                    try {
                        String getterName = "get"+setterName.substring(3);
                        Method getter =
                            cldst.getMethod(getterName,new Class[]{});
                        if( !getter.getReturnType().equals(params[0])) {
                            // The getter hasn't been found
                            throw new NoSuchMethodException();
                        }
                        // The getter has been found. This is a read-write
                        // attribute. It must have been taken into account by
                        // initializeReadAttributes().
                    }
                    catch(NoSuchMethodException nsme) {
                        // No associated getter
                        try {
                            Field field = getField(cl,setter);
                            wAttributes.put(setter,field);
                        }
                        catch( NoSuchFieldException nsfe ) {
                            // No associated field.
                            // Can't do anything.
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Attributes with a getter Method and a setter Method.
     * key=getter Method. value=associated setter Method.
     */
    private Map rwAttributes;
    
    /**
     * Read-only attributes with only a getter Method.
     * key=getter Method. value=associated Field.
     */
    private Map rAttributes;
    
    /**
     * Write-only attributes with only a setter Method.
     * key=setter Method. value=associated Field.
     */
    private Map wAttributes;    
}
