/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.control.binding;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.control.component.ComponentItf;
import aokell.lib.membrane.marker.BindingType;
import aokell.lib.type.InterfaceTypeImpl;
import aokell.lib.util.FeatureHelper;
import aokell.lib.util.FractalHelper;

/**
 * Root class for binding controllers implementations.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public abstract class AbstractBindingControllerImpl
    implements BindingControllerItf, Controller {
    
    /**
     * The bindings for this component.
     * key=client interface name.
     * value=bound server interface.
     */
    private Map bindings = new HashMap();
    
    public AbstractBindingControllerImpl() {}
    
    
    // --------------------------------------------------------------
    // BindingController interface
    // --------------------------------------------------------------
    
    /**
     * Returns the names of the client interfaces of the component to which this
     * interface belongs.
     *
     * @return the names of the client interfaces of the component to which this
     *      interface belongs.
     */
    public String[] listFc() {
        
        /*
         * For Julia-like components, notify the content. AspectJ-like
         * components can not be notified: the control interface is directly
         * introduced into the content.
         */
        Object contentPart = FractalHelper.getContent(compctrl);
        if( ! (contentPart instanceof BindingType) &&
            contentPart instanceof BindingController ) {
            ((BindingController)contentPart).listFc();
        }
        
        // Get client singleton interfaces
        List cltItfs = getFcSingletonClientItfs();
        List cltItfNames = new ArrayList( cltItfs.size() );
        
        for (Iterator iter = cltItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            cltItfNames.add( itf.getFcItfName() );
        }
        
        // Add collection client interfaces currently bound
        addFcCollectionInterfaceNames( cltItfNames );
        
        return (String[]) cltItfNames.toArray(new String[cltItfNames.size()]);        
    }

    /**
     * Returns the interface to which the given client interface is bound. More
     * precisely, returns the server interface to which the client interface
     * whose name is given is bound. This server interface is necessarily in
     * the same address space as the client interface.
     *
     * @param arg0 the name of a client interface of the component to
     *      which this interface belongs.
     * @return the server interface to which the given interface is bound, or
     *      <tt>null</tt> if it is not bound.
     * @throws NoSuchInterfaceException if the component to which this interface
     *      belongs does not have a client interface whose name is equal to the
     *      given name.
     */
    public Object lookupFc(String arg0) throws NoSuchInterfaceException {
        
        /*
         * For Julia-like components, notify the content. AspectJ-like
         * components can not be notified: the control interface is directly
         * introduced into the content.
         */
        Object contentPart = FractalHelper.getContent(compctrl);
        if( ! (contentPart instanceof BindingType) &&
            contentPart instanceof BindingController ) {
            try {
                ((BindingController)contentPart).lookupFc(arg0);
            }
            // Ignore exceptions which may be thrown by the content
            catch( NoSuchInterfaceException nsie ) {}
        }
        
        if( arg0.equals("component") ) {
            // Do not perform any check for the interface "component"
            return bindings.get(arg0);
        }
        
        Object bound = bindings.get(arg0);

        // Bindings for collection interfaces are registered with the name of
        // the interface followed by a suffix (0, 1, 2, ...). Hence looking up
        // only the interface name returns null. We then return as a List
        // all the interfaces whose binding name starts with the prefix.
        
        if( bound==null ) {
            
            // throw NoSuchInterfaceException if the interface does not exist
            Interface itf = (Interface) getFcClientItf(arg0);            
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            
            if( it.isFcCollectionItf() ) {
                Set sorted = new TreeSet(bindings.keySet());
                List bounds = null;
                for (Iterator iter = sorted.iterator(); iter.hasNext();) {
                    String element = (String) iter.next();
                    if( element.startsWith(arg0) ) {
                        if( bounds == null )
                            bounds = new ArrayList();
                        bounds.add( bindings.get(element) );
                    }
                }
                return bounds;
            }
        }
        
        return bound;
    }

    /**
     * Binds the client interface whose name is given to a server interface.
     * More precisely, binds the client interface of the component to which this
     * interface belongs, and whose name is equal to the given name, to the
     * given server interface. The given server interface must be in the same
     * address space as the client interface.
     *
     * @param arg0 the name of a client interface of the component to
     *      which this interface belongs.
     * @param arg1 a server interface.
     * @throws NoSuchInterfaceException if there is no such client interface.
     * @throws IllegalBindingException if the binding cannot be created.
     * @throws IllegalLifeCycleException if this component has a {@link
     *      LifeCycleController} interface, but it is not in an appropriate
     *      state to perform this operation.
     */
    public void bindFc(String arg0, Object arg1)
    throws NoSuchInterfaceException, IllegalBindingException,
    IllegalLifeCycleException {
        
        bindFcContent(arg0,arg1);

        if( arg0.equals("component") ) {
            // Do not perform any check for the interface "component"
            bindings.put(arg0,arg1);
            return;
        }
        
        // throw NoSuchInterfaceException if the interface does not exist
        Interface itf = (Interface) getFcClientItf(arg0);
        InterfaceType it = (InterfaceType) itf.getFcItfType();
        
        // Test whether the binding is legal
        testLegalBinding(arg0,it,arg1);
        
        // Test whether the interface is not already bound
        if( bindings.containsKey(arg0) )
            throw new IllegalBindingException(
                    "Interface "+arg0+" already bound");
        
        // All tests cleared. Record the binding.
        bindings.put(arg0,arg1);
        
        // Record the reverse binding
        Interface dstItf = (Interface) arg1;
        Component dstItfOwner = dstItf.getFcItfOwner();
        BindingControllerItf bc =
            FractalHelper.getBindingController(dstItfOwner);
        bc.reverseBindFc( dstItf.getFcItfName(), itf );
    }
    
    /**
     * If the content and the component parts are separated, and if the
     * content part implements the BindingController interface,
     * invoke it. Discard the result.
     */
    protected void bindFcContent(String arg0, Object arg1) {
        
        /*
         * For Julia-like components, notify the content. AspectJ-like
         * components can not be notified: the control interface is directly
         * introduced into the content.
         */
        Object contentPart = FractalHelper.getContent(compctrl);
        if( ! (contentPart instanceof BindingType) &&
            contentPart instanceof BindingController ) {
            try {
                ((BindingController)contentPart).bindFc(arg0,arg1);
            }
            // Ignore exceptions which may be thrown by the content
            catch( NoSuchInterfaceException nsie ) {}
            catch( IllegalBindingException ibe ) {}
            catch( IllegalLifeCycleException ilce ) {}
        }
    }
    
    /**
     * Unbinds the given client interface. More precisely, unbinds the client
     * interface of the component to which this interface belongs, and whose
     * name is equal to the given name.
     *
     * @param arg0 the name of a client interface of the component to
     *      which this interface belongs.
     * @throws NoSuchInterfaceException if there is no such client interface.
     * @throws IllegalBindingException if the binding cannot be removed.
     * @throws IllegalLifeCycleException if this component has a {@link
     *      LifeCycleController} interface, but it is not in an appropriate
     *      state to perform this operation.
     */
    public void unbindFc(String arg0)
    throws NoSuchInterfaceException, IllegalBindingException,
    IllegalLifeCycleException {
        
        Component compctrl = FeatureHelper.getFcImplementation(this.compctrl);
        
        /*
         * For Julia-like components, notify the content. AspectJ-like
         * components can not be notified: the control interface is directly
         * introduced into the content.
         */
        Object contentPart = FractalHelper.getContent(compctrl);
        if( ! (contentPart instanceof BindingType) &&
            contentPart instanceof BindingController ) {
            try {
                ((BindingController)contentPart).unbindFc(arg0);
            }
            // Ignore exceptions which may be thrown by the content
            catch( NoSuchInterfaceException nsie ) {}
            catch( IllegalBindingException ibe ) {}
            catch( IllegalLifeCycleException ilce ) {}
        }

        // getFcInterfaceType() may throw NoSuchInterfaceException
        ComponentType ct = (ComponentType) compctrl.getFcType();
        ct.getFcInterfaceType(arg0);
        
        // Test whether the interface is bound
        if( !bindings.containsKey(arg0) )
            throw new IllegalBindingException(
                    "Interface "+arg0+" not bound");            
        
        // Test whether the component is stopped
        testLifeCyleBeforeUnbinding(compctrl);
        
        // Remove the bindings
        Object arg1 = bindings.remove(arg0);
        
        // Remove the reverse binding
        Interface dstItf = (Interface) arg1;
        Component dstItfOwner = dstItf.getFcItfOwner();
        BindingControllerItf bc = FractalHelper.getBindingController(dstItfOwner);
        bc.reverseUnbindFc( dstItf.getFcItfName() );
    }

    
    // --------------------------------------------------------------
    // BindingControllerItf interface
    // --------------------------------------------------------------
    
    /**
     * Store backward links to interfaces bound to the interfaces of the
     * component managed by this binding controller.
     */
    public Map reverseBindings = new HashMap();
    
    /**
     * Record a reverse binding. When a source interface is bound to a target
     * one, the reverse link (target -&gt; source) is also recorded in the
     * target binding controller. This reverse mappings is primarily used to
     * prevent content controllers from performing remove operations that would
     * create invalid bindings.
     * 
     * @param dstItfName  the target interface name for the binding
     * @param src         the source interface for the binding
     */
    public void reverseBindFc( String dstItfName, Object src ) {
        reverseBindings.put(dstItfName,src);
    }
    
    /**
     * Lookup a reverse binding. See reverseBindFc(String,Object) for the
     * rationale behind reverse bindings.
     * 
     * @param dstItfName  the target interface name to be looked up
     * @return            the corresponding source interface or null
     */
    public Object reverseLookupFc( String dstItfName ) {
        return reverseBindings.get(dstItfName);
    }
    
    /**
     * Remove a reverse binding. See reverseBindFc(String,Object) for the
     * rationale behind reverse bindings.
     * 
     * @param dstItfName  the target interface name to be removed
     */
    public void reverseUnbindFc( String dstItfName ) {
        reverseBindings.remove(dstItfName);
    }
    
    /**
     * @return  true if the given component contains at least one binding
     * (excluding the binding with the component interface).
     */
    public boolean isFcBound() {
        
        for (Iterator iter = bindings.keySet().iterator(); iter.hasNext();) {
            String itfName = (String) iter.next();
            
            // Exclude bindings towards the component interface
            if( ! itfName.equals("component") )
                return true;
        }
        
        return false;
    }
    
    /**
     * @return  true if the given component contains at least one reverse
     * binding
     */
    public boolean isFcReverseBound() {        
        return ! reverseBindings.isEmpty();
    }
    
    /**
     * Add the names of currently bound interfaces to the given list.
     */
    public void addFcCollectionInterfaceNames( List cltItfNames ) {
        
        ComponentItf c =
            (ComponentItf) FeatureHelper.getFcImplementation(compctrl);
        List collectionItfs = c.getFcCollectionInterfaces();
        
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            for (Iterator iter2 = bindings.keySet().iterator();
                    iter2.hasNext();) {
                String key = (String) iter2.next();
                if( key.startsWith(itf.getFcItfName()) ) {
                    cltItfNames.add(key);
                }
            }
        }
    }
    

    // --------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------
    
    protected void testLifeCyleBeforeUnbinding( Component arg0 )
    throws IllegalLifeCycleException {
        
        LifeCycleController lc = FractalHelper.getLifeCycleController(arg0);
        if( lc.getFcState().equals(LifeCycleController.STARTED) ) {
            throw new IllegalLifeCycleException(
                    "Component must be stopped before performing an operation "+
                    "on the binding controller" );
        }
    }
    
    /**
     * Return the interface whose name is given.
     */
    protected Object getFcClientItf(String itfName)
        throws NoSuchInterfaceException {
        
        ComponentItf c =
            (ComponentItf) FeatureHelper.getFcImplementation(compctrl);
        
        // Lookup in client external interfaces
        return c.getFcClientInterface(itfName);
    }
    
    /**
     * Return the list of client interfaces for the component attached to this
     * controller.
     */
    protected List getFcSingletonClientItfs() {
        
        List clientItfs = new ArrayList();
        
        ComponentItf c =
            (ComponentItf) FeatureHelper.getFcImplementation(compctrl);

        // Add client external interfaces
        List sbi = c.getFcSingletonClientInterfaces();
        clientItfs.addAll(sbi);
        
        return clientItfs;
    }
    
    /**
     * @return    return true if it1 is assignable from it2
     */
    private static boolean isAssignableFrom(
            InterfaceType it1, InterfaceType it2 ) {
        
        Class i1,i2;
        try {
            i1 = ((InterfaceTypeImpl)it1).getFcItfClass();
            i2 = ((InterfaceTypeImpl)it2).getFcItfClass();
        }
        catch(ClassNotFoundException cnfe) {
            return false;
        }

        return i1.isAssignableFrom(i2);
    }
    
    /**
     * Test whether the a client interface implementing the specified
     * InterfaceType can be bound to arg1.
     * 
     * @throws IllegalBindingException  if the binding is illegal
     */
    protected void testLegalBinding(
            String arg0, InterfaceType it, Object arg1 )
    throws IllegalBindingException, NoSuchInterfaceException {
        
        Interface targetItf = (Interface) arg1;
        InterfaceType targetItfType = (InterfaceType) targetItf.getFcItfType();
        String targetItfName = targetItf.getFcItfName();
        
        if( (! it.isFcOptionalItf()) && targetItfType.isFcOptionalItf() )
            throw new IllegalBindingException(
                    "Mandatory to optional bindings are forbidden: "+
                    "("+arg0+","+it+")->("+targetItfType+","+arg1);

        if( ! isAssignableFrom(it,targetItfType) )
            throw new IllegalBindingException(
                    "Target interface "+targetItfName+
                    " is not a sub-type of source interface "+
                    arg0+" type");
        
//        // Test for export bindings
//        if( ContentControllerImpl.isFcContainedIn(target,component) ) {
//            if( it.isFcClientItf() ) {
//                throw new IllegalBindingException(it+" is client");
//            }
//            if( targetItfType.isFcClientItf() ) {
//                throw new IllegalBindingException(targetItfType+" is client");
//            }
//        }        
//        // Test for import bindings
//        else if( ContentControllerImpl.isFcContainedIn(component,target) ) {
//            if( ! it.isFcClientItf() ) {
//                throw new IllegalBindingException(it+" is server");
//            }
//            if( ( targetItf.isFcInternalItf() &&
//                  targetItfType.isFcClientItf() )
//                ||
//                ( ! targetItf.isFcInternalItf() &&
//                  ! targetItfType.isFcClientItf() ) ) {
//                throw new IllegalBindingException(targetItfType+" is server");
//            }
//        }        
//        // Test for normal bindings
//        else {
            if( ! it.isFcClientItf() ) {
                throw new IllegalBindingException(it+" is server");
            }
            if( targetItfType.isFcClientItf() ) {
                throw new IllegalBindingException(targetItfType+" is client");
            }
//        }
                
    }
    
    
    // --------------------------------------------------------------
    // Controller implementation
    // --------------------------------------------------------------
    
    protected Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        
        /*
         * Initialize the binding towards the component interface.
         */
        try {
            bindFc("component",compctrl);
        }
        catch( Exception e ) {
            throw new RuntimeException(e);
        }

        /*
         * Optimization: direct reference to the binding controller.
         */
        Object content = FractalHelper.getContent(compctrl);
        
        if( content instanceof BindingControllerSetterGetterItf ) {
            ((BindingControllerSetterGetterItf) content).
                setFcBindingController(this);
        }
    }
    
    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {
        
        /*
         * Map storing associations between the current component (plus its
         * content if this is a composite component) and their associated newly
         * created component.
         */
        Map newContent = (Map) hints;
        BindingController newBc = FractalHelper.getBindingController(dst);
        
        for (Iterator iter = bindings.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            String srcItfName = (String) entry.getKey();
            
            if( ! srcItfName.equals("component") ) {
                Interface dstItf = (Interface) entry.getValue();
                try {
                    cloneBinding(srcItfName,dstItf,newContent,newBc);
                }
                catch(Exception e) {
                    e.printStackTrace();
                    throw new CloneCtrlException(e);
                }
            }
        }
    }
    
    /**
     * Clone a binding.
     * 
     * @param itfName     the source interface name
     * @param boundItf    the interface bound to the source
     * @param newContent  the map associating old components and new ones
     * @param newBc       the binding controller of the new component
     */
    private void cloneBinding(
            String itfName, Interface boundItf,
            Map newContent, BindingController newBc)
    throws Exception {
        
        if( boundItf == null )
            return;
        
        String boundItfName = boundItf.getFcItfName();        
        Component boundCompCtrl = boundItf.getFcItfOwner();
        
        Component dstBound = (Component) newContent.get(boundCompCtrl);
        
        if( dstBound == null ) {
            /*
             * Trying to clone a binding starting from the cloned components
             * and targeting a component which is outside the scope of the
             * cloned components.
             */
            return;
        }
        
        // Check whether the old binding was toward an internal or an external
        // interface
        Object dstBoundItf = null;
        if( boundItf.isFcInternalItf() ) {
            ContentController cc = FractalHelper.getContentController(dstBound);
            dstBoundItf = cc.getFcInternalInterface(boundItfName);
        }
        else {
            dstBoundItf = dstBound.getFcInterface(boundItfName);
        }
        
        // Record the new binding
        newBc.bindFc(itfName,dstBoundItf);        
    }
}
