/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.control.component;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.lib.CloneableItf;
import aokell.lib.InterfaceImpl;
import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.control.binding.BindingControllerItf;
import aokell.lib.interf.Delegator;
import aokell.lib.interf.DelegatorForBoundableInterface;
import aokell.lib.interf.DelegatorForImplementedInterface;
import aokell.lib.util.FractalHelper;

/**
 * Root class for implementations of the Component interface.
 * This class contains common features shared by all classes implementing this
 * interface.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class ComponentImpl
    
    /*
     * Do not directly implements Interface as this would lead to a behavior
     * under fractal-explorer which is different from the one obtained with
     * Julia. Extending InterfaceImpl is better.
     */
    extends InterfaceImpl
    
    implements ComponentItf, ComponentDescItf, Controller {
    
    private Type type;
    private Object controllerDesc;
    private Object contentDesc;

    public ComponentImpl() {
        /*
         * Can not call super as the 1st argument should be this (and this is
         * forbidden when calling super). See below: Interface is overriden.
         */
    }
    
    
    // ------------------------------------------------------------------
    // Implementation of the Component interface
    // ------------------------------------------------------------------
    
    public Type getFcType() {
        return type;
    }

    /**
     * @return  the instance implementing the given interface name for this
     * component
     */
    public Object getFcInterface(String interfaceName)
    throws NoSuchInterfaceException {
        
        // Lookup the name in the singleton Interface list
        List singletonItfs = getFcSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        // Lookup the name in the collection Interface list
        List collectionItfs = getFcCollectionInterfaces();
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        /*
         * The interface has not been found in the collection interface list.
         * A possible cause if that a interface name such as foo004 is looked
         * up. In such a case the corresponding Interface instance must be
         * generated on the fly, and registered into the list of collection
         * interfaces.
         */
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( interfaceName.startsWith(itf.getFcItfName()) ) {
                
                /*
                 * Check whether the requested interface has been requested
                 * before.
                 */
                List dynCollectionItfs = getFcDynCollectionInterfaces();
                for (Iterator iterator = dynCollectionItfs.iterator(); iterator
                        .hasNext();) {
                    Interface element = (Interface) iterator.next();
                    if( interfaceName.equals(element.getFcItfName()) ) {
                        return element;
                    }
                }
                
                /*
                 * Generate a new interface.
                 */
                InterfaceImpl clone = (InterfaceImpl) ((CloneableItf)itf).clone();
                clone.setFcItfName(interfaceName);
                dynCollectionItfs.add(clone);
                
                return clone;
            }            
        }
        
        /*
         * Components with a content (all primitives have a content, composites
         * may have a content) implements the /content interface.
         * The /content interface is not defined in the Fractal specifications
         * but is Julia-specific.
         */
        if( interfaceName.equals("/content") ) {
            return getContentPart();
        }
        
        if( interfaceName.equals("/desc") ) {
            return this;
        }
        
        /*
         * This was our last chance. If the interface has still not been found,
         * throw NoSuchInterfaceException.
         */
        
        throw new NoSuchInterfaceException(interfaceName);
    }
    
    /**
     * The array of Interface instances provided by this component and
     * returned by getFcInterfaces().
     */
    private Object[] itfs;
    
    /**
     * Return the array of Interface instances implemented by this component.
     */
    public Object[] getFcInterfaces() {
        
        if( itfs == null ) {
            
            // Get singleton interfaces
            List singletonItfs = getFcSingletonInterfaces();
            
            // Remove interfaces with a name which begins with /
            List sItfs = new ArrayList(singletonItfs);
            for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
                Interface itf = (Interface) iter.next();
                if( itf.getFcItfName().startsWith("/") ) {
                    sItfs.remove(itf);
                }
            }
            
            itfs = (Interface[]) sItfs.toArray( new Interface[sItfs.size()] );
        }
        
        // Add client collection interfaces currently bound
        List colItfs = new ArrayList();
        
        BindingControllerItf bc = FractalHelper.getBindingController(this);
        
        List collectionInterfaces = getFcCollectionInterfaces();
        List boundedCollectionInterfaceNames = new ArrayList();
        bc.addFcCollectionInterfaceNames(boundedCollectionInterfaceNames);
        
        for (Iterator iter = boundedCollectionInterfaceNames.iterator(); iter
                .hasNext();) {
            String bciName = (String) iter.next();
            for (Iterator iterator = collectionInterfaces.iterator(); iterator
                    .hasNext();) {
                Interface itf = (Interface) iterator.next();
                if( bciName.startsWith(itf.getFcItfName()) ) {
                    /*
                     * The collection Interface instance corresponding to the
                     * bound name has been found.
                     */
                    InterfaceImpl clone =
                        (InterfaceImpl) ((CloneableItf)itf).clone();
                    clone.setFcItfName(bciName);
                    colItfs.add(clone);
                }
            }
        }
        
        // Add server collection interfaces currently bound
        for (Iterator iter = collectionInterfaces.iterator(); iter
                .hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( ! it.isFcClientItf() ) {
                String itfName = itf.getFcItfName();
                if( bc.reverseLookupFc(itfName) != null ) {
                    // The server collection interface is currently bound
                    colItfs.add(itf);
                }
            }
        }
        
        
        if( colItfs.size() == 0 ) {
            // No collection interface bound
            return itfs;
        }
        
        colItfs.addAll( Arrays.asList(itfs) );
        return colItfs.toArray();
    }
    
    
    // ------------------------------------------------------------------
    // Implementation of the ComponentDescItf interface
    // ------------------------------------------------------------------
        
    public Object getControllerDesc() {
        return controllerDesc;
    }
    
    public Object getContentDesc() {
        return contentDesc;
    }
 
    
    // ------------------------------------------------------------------
    // Implementation of the ComponentItf interface
    // ------------------------------------------------------------------
        
    public void setType(Type type) {
        this.type = type;
    }

    public void setContentDesc(Object contentDesc) {
        this.contentDesc = contentDesc;
    }

    public void setControllerDesc(Object controllerDesc) {
        this.controllerDesc = controllerDesc;
    }
    
    // <InterfaceType,Controller>
    private Map ctrls = new HashMap();
    
    /**
     * Return the array of control interfaces implemented by the component.
     */
    public Controller[] getFcControllers() {
        Collection values = ctrls.values();
        return (Controller[]) values.toArray( new Controller[values.size()] );
    }

    /**
     * Declare a new controller implemented by the component.
     */
    public void addFcController( InterfaceType it, Controller ctrl ) {
        ctrls.put(it,ctrl);
    }
    
    /**
     * Return the list of client interfaces implemented by this component.
     */
    public List getFcSingletonClientInterfaces() {
        
        List clientItfs = new ArrayList();
        
        // Add client singleton interfaces
        List singletonItfs = getFcSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() )
                clientItfs.add(itf);
        }
        
//        // Add client collection interfaces
//        List collectionItfs = getFcCollectionInterfaces();
//        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
//            Interface itf = (Interface) iter.next();
//            InterfaceType it = (InterfaceType) itf.getFcItfType();
//            if( it.isFcClientItf() )
//                clientItfs.add(itf);
//        }
        
        return clientItfs;
    }
    
    /**
     * Return the instance implementing the given client interface name for this
     * component. This method is a clone for getFcInterface(String).
     */
    public Object getFcClientInterface(String interfaceName)
    throws NoSuchInterfaceException {
        
        // Lookup the name in the singleton Interface list
        List singletonItfs = getFcSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() &&
                itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        /*
         * Lookup the name in the collection Interface list.
         * When such an interface is looked up its name is something like foo004
         * which does not exists as an interface name (but foo does).
         * See TestBindingController.
         */
        List collectionItfs = getFcCollectionInterfaces();
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() &&
                interfaceName.startsWith(itf.getFcItfName()) ) {
                return itf;
            }            
        }
        
        throw new NoSuchInterfaceException(interfaceName);
    }
    
    /** The list of Interface instances for singleton interfaces. */
    private List singletonItfs;
    
    /**
     * @return  the list of singleton Interface instances implemented by this
     *          component
     */
    public List getFcSingletonInterfaces() {
        
        if( singletonItfs != null )
            return singletonItfs;
        
        List sbi = getFcSingletonBusinessInterfaces();
        List sci = getFcSingletonControlInterfaces();
        
        singletonItfs = new ArrayList( sbi.size() + sci.size() );
        singletonItfs.addAll(sbi);
        singletonItfs.addAll(sci);
        
        return singletonItfs;
    }
    
    /** The list of Interface instances for singleton business interfaces. */
    private List singletonBusinessItfs;
    
    /**
     * @return  the list of singleton business Interface instances implemented
     *          by this component
     */
    public List getFcSingletonBusinessInterfaces() {
        
        if( singletonBusinessItfs != null )
            return singletonBusinessItfs;
        
        singletonBusinessItfs = new ArrayList();
        
        ComponentType ct = (ComponentType) getFcType();
        InterfaceType[] its = ct.getFcInterfaceTypes();
        
        // User-defined singleton interfaces 
        for (int i = 0; i < its.length; i++) {
            if( ! its[i].isFcCollectionItf() ) {
                Interface itf = getFcInterface(its[i],this);
                singletonBusinessItfs.add(itf);
            }
        }
        
        return singletonBusinessItfs;
    }
    
    /** The list of Interface instances for singleton control interfaces. */
    private List singletonControlItfs;
    
    /**
     * @return  the list of singleton control Interface instances implemented
     *          by this component
     */
    public List getFcSingletonControlInterfaces() {
        
        if( singletonControlItfs != null )
            return singletonControlItfs;
        
        singletonControlItfs = new ArrayList();
        
        int i = 0;
        Delegator dfii = DelegatorForImplementedInterface.get();
        
        for (Iterator iter = ctrls.entrySet().iterator(); iter.hasNext(); i++) {
            Map.Entry entry = (Map.Entry) iter.next();
            InterfaceType it = (InterfaceType) entry.getKey();
            Controller ctrl = (Controller) entry.getValue();
            
            Interface itf = getProxyInterface(it,this,ctrl,dfii);
            singletonControlItfs.add(itf);
        }
        
        return singletonControlItfs;
    }
    
    /** The list of Interface instances for collection interfaces. */
    private List collectionItfs;
    
    /**
     * @return  the list of collection Interface instances implemented by this
     *          component
     */
    public List getFcCollectionInterfaces() {
        
        if( collectionItfs != null )
            return collectionItfs;
        
        collectionItfs = new ArrayList();
        
        ComponentType ct = (ComponentType) getFcType();
        InterfaceType[] its = ct.getFcInterfaceTypes();
        
        // User-defined collection interfaces 
        for (int i = 0; i < its.length; i++) {
            if( its[i].isFcCollectionItf() ) {
                Interface itf = getFcInterface(its[i],this);
                collectionItfs.add(itf);
            }
        }
        
        return collectionItfs;
    }
    
    /**
     * The list of dynamically defined collection Interface instances (e.g. foo004).
     */
    private List dynCollectionItfs;
    
    /**
     * @return  the list of dynamically defined collection Interface instances
     *          (e.g. foo004).
     */
    private List getFcDynCollectionInterfaces() {
        if( dynCollectionItfs == null ) {
            dynCollectionItfs = new ArrayList();
        }
        
        return dynCollectionItfs;
    }
    
   /**
     * Get the Interface instance associated to the given interface type for the
     * given component.
     */
    protected Interface getFcInterface( InterfaceType it, Component itfOwner ) {
        
        Delegator dfii = DelegatorForImplementedInterface.get();
        Delegator dfbi = DelegatorForBoundableInterface.get();
        return
            ( it.isFcClientItf() ) ?
            getProxyInterface(it,itfOwner,getContentPart(),dfbi) :
            getProxyInterface(it,itfOwner,getContentPart(),dfii);
    }
    
    /**
     * Return the Interface instance associated to the given interface type.
     * The returned instance implements both the Interface interface and the
     * Java interface specified in the given interface type. The instance is a
     * proxy generated with ASM.
     * 
     * @param it         the interface type
     * @param itfOwner   the component owning this interface
     * @param content    the instance implementing the content
     * @param delegator  the class used to generate the proxy methods
     * @return           the interface
     */
    public static Interface getProxyInterface(
            InterfaceType it, Component itfOwner, Object content,
            Delegator delegator ) {
        
        Object proxy = delegator.generate(
                it,                 // interface type              
                itfOwner,           // component owning the interface
                content,            // the instance holding the content
                false               // external interface
                );          
        
        return (Interface) proxy;
    }

    private Object contentPart;
    
    public void setContentPart( Object contentPart ) {
        this.contentPart = contentPart;
    }
    
    protected Object getContentPart() {
        return contentPart;
    }
    
    
    // --------------------------------------------------------------
    // Interface interface
    // --------------------------------------------------------------
    
    public String getFcItfName() {
        return ComponentItf.NAME;
    }
    
    public Component getFcItfOwner() {
        return this;
    }
    
    public Type getFcItfType() {
        return ComponentItf.TYPE;
    }
    
    public boolean isFcInternalItf() {
        return false;
    }

    
    // --------------------------------------------------------------
    // Controller interface
    // --------------------------------------------------------------
    
    protected Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     * 
     * @param compctrl  the component controller associated with this controller
     * For instances of this class, compctrl and this are equals (the component
     * controller is associated with itself).
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        
        // (re)initialize cached values
        collectionItfs = null;
        dynCollectionItfs = null;
        itfs = null;
        singletonBusinessItfs = null;
        singletonControlItfs = null;
        singletonItfs = null;
    }
    
    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {
        // Indeed nothing for the component controller
    }

}
