/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.control.content;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.lib.CloneableItf;
import aokell.lib.InterfaceImpl;
import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.control.binding.BindingControllerItf;
import aokell.lib.control.component.ComponentItf;
import aokell.lib.control.superc.SuperControllerItf;
import aokell.lib.type.InterfaceTypeImpl;
import aokell.lib.util.FeatureHelper;
import aokell.lib.util.FractalHelper;

/**
 * Implementation for the content controller.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class ContentControllerImpl
    implements ContentControllerItf, Controller {
    
    /**
     * The list of components under the control of this controller.
     * The type should be List<Component>.
     */
    private List content = new ArrayList();

    public ContentControllerImpl() {}
    
    
    // --------------------------------------------------------------
    // ContentController interface
    // --------------------------------------------------------------
    
    private Object[] internalInterfacesArray;
    
    /**
     * @return  the array of internal Interface instances implemented by this
     * component
     */
    public Object[] getFcInternalInterfaces() {
        if( internalInterfacesArray == null ) {
            List singletonItfs = getFcInternalSingletonInterfaces();
            internalInterfacesArray =
                singletonItfs.toArray(new Object[singletonItfs.size()]);
        }
        
        // TODO collection interfaces currently bound (see Component.getFcInterfaces())
        
        return internalInterfacesArray;
    }

    public Object getFcInternalInterface(String interfaceName)
    throws NoSuchInterfaceException {
        
        // Lookup the name in the singleton Interface list
        List singletonItfs = getFcInternalSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        // Lookup the name in the collection Interface list
        List collectionItfs = getFcInternalCollectionInterfaces();
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        /*
         * The interface has not been found in the collection interface list.
         * A possible cause if that a interface name such as foo004 is looked
         * up. In such a case the corresponding Interface instance must be
         * generated on the fly, and registered into the list of collection
         * interfaces.
         */
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            if( interfaceName.startsWith(itf.getFcItfName()) ) {
                InterfaceImpl clone = (InterfaceImpl) ((CloneableItf)itf).clone();
                clone.setFcItfName(interfaceName);
                collectionItfs.add(clone);
                return clone;
            }            
        }
        
        throw new NoSuchInterfaceException(interfaceName);
    }

    public Component[] getFcSubComponents() {
        return (Component[]) content.toArray(new Component[content.size()]);
    }

    public void addFcSubComponent(Component arg0)
    throws IllegalContentException, IllegalLifeCycleException {
        
        Component compctrl = FeatureHelper.getFcImplementation(this.compctrl);
        
        // Test on the containment hierarchy
        
        // 1st test: is the added one (arg0) already contained in the comp?
        // Not recursive because of a problem with shared when a component is
        // shared by the root and a component that is contained in the root
        // (somewhat tricky but seems to be permitted by Julia and Nicolas uses
        // it)
        if( isFcContainedIn(false,arg0,compctrl) ) {
            throw new IllegalContentException(
                    arg0+" is already a sub-component of "+compctrl);
        }
        
        // 2nd test: is the component already contained in the added one?
        // Recursive to detect non trivial cycles.
        if( arg0==compctrl || isFcContainedIn(true,compctrl,arg0) ) {
            throw new IllegalContentException(
                    "Adding "+arg0+" in "+compctrl+" would create a cycle");
        }
        
        // Add the sub-component
        content.add(arg0);
        
        // Update the super controller state of the sub-component
        SuperControllerItf sc = FractalHelper.getSuperController(arg0);
        sc.addFcSuperComponent(compctrl);
    }

    public void removeFcSubComponent(Component arg0)
    throws IllegalContentException, IllegalLifeCycleException {
        
        Component compctrl = FeatureHelper.getFcImplementation(this.compctrl);
        
        // Test on the containment hierarchy
        if( ! content.contains(arg0) ) {
            throw new IllegalContentException(
                    arg0+" is not a sub-component of "+compctrl);
        }
        
        // Test on bindings
        BindingControllerItf bc = FractalHelper.getBindingController(arg0);
        boolean isBound = bc.isFcBound();
        boolean isRBound = bc.isFcReverseBound();
        if( isBound || isRBound ) {
            throw new IllegalContentException(
                    "Removing "+arg0+" from "+compctrl+
                    " would create illegal bindings");
        }
        
        // Test whether the removed component is stopped
        testLifeCyleBeforeRemoving(arg0);
        
        // Remove the sub-component
        content.remove(arg0);
        
        // Update the super controller state of the sub-component
        SuperControllerItf sc = FractalHelper.getSuperController(arg0);
        sc.removeFcSuperComponent(compctrl);        
    }

    
    // --------------------------------------------------------------
    // ContentControllerItf interface
    // --------------------------------------------------------------
    
    /**
     * Return the list of client internal interfaces.
     */
    public List getFcInternalSingletonClientInterfaces() {
        
        List clientItfs = new ArrayList();
        
        // Add client singleton internal interfaces
        List singletonItfs = getFcInternalSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() ) {
                clientItfs.add(itf);
            }            
        }
        
//        // Add client collection internal interfaces
//        List collectionItfs = getFcInternalCollectionInterfaces();
//        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
//            Interface itf = (Interface) iter.next();
//            InterfaceType it = (InterfaceType) itf.getFcItfType();
//            if( it.isFcClientItf() ) {
//                clientItfs.add(itf);
//            }            
//        }
        
        return clientItfs;
    }

    /**
     * Return the instance implementing the given client internal interface
     * name. This method is a clone for getFcInternalInterface(String).
     */
    public Object getFcInternalClientInterface(String interfaceName)
    throws NoSuchInterfaceException {
        
        // Lookup the name in the singleton Interface list
        List singletonItfs = getFcInternalSingletonInterfaces();
        for (Iterator iter = singletonItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() &&
                itf.getFcItfName().equals(interfaceName) ) {
                return itf;
            }            
        }
        
        /*
         * Lookup the name in the collection Interface list.
         * When such an interface is looked up its name is something like foo004
         * which does not exists as an interface name (but foo does).
         * See TestBindingController.
         */
        List collectionItfs = getFcInternalCollectionInterfaces();
        for (Iterator iter = collectionItfs.iterator(); iter.hasNext();) {
            Interface itf = (Interface) iter.next();
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() &&
                interfaceName.startsWith(itf.getFcItfName()) ) {
                return itf;
            }            
        }
        
        throw new NoSuchInterfaceException(interfaceName);
    }

    
    // --------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------
    
    protected void testLifeCyleBeforeRemoving( Component arg0 )
    throws IllegalLifeCycleException {
        
        LifeCycleController lc = FractalHelper.getLifeCycleController(arg0);
        if( lc.getFcState().equals(LifeCycleController.STARTED) )
            throw new IllegalLifeCycleException(
                    "Component "+arg0+" must be stopped before removed");
    }
    
    private List singletonInterfaces;
    
    /**
     * @return  the list of internal singleton Interface instances implemented
     * by this component
     */
    private List getFcInternalSingletonInterfaces() {
        if( singletonInterfaces == null ) {
            List isbi = getFcInternalSingletonBusinessInterfaces();
            List isci = getFcInternalSingletonControlInterfaces();
            singletonInterfaces = new ArrayList( isbi.size() + isci.size() );
            singletonInterfaces.addAll( isbi );
            singletonInterfaces.addAll( isci );
        }
        return singletonInterfaces;        
    }

    private List singletonBusinessInterfaces;
    
    /**
     * @return  the list of internal singleton business Interface instances 
     * implemented by this component
     */
    private List getFcInternalSingletonBusinessInterfaces() {
        if( singletonBusinessInterfaces == null ) {
            ComponentItf c =
                (ComponentItf) FeatureHelper.getFcImplementation(compctrl);
            List itfs = c.getFcSingletonBusinessInterfaces();
            singletonBusinessInterfaces = getFcInternalInterfaces(itfs);
        }
        return singletonBusinessInterfaces;        
    }

    private List singletonControlInterfaces;
    
    /**
     * @return  the list of internal singleton control Interface instances 
     * implemented by this component
     */
    private List getFcInternalSingletonControlInterfaces() {
        if( singletonControlInterfaces == null ) {
            ComponentItf c =
                (ComponentItf) FeatureHelper.getFcImplementation(compctrl);
            List itfs = c.getFcSingletonControlInterfaces();
            singletonControlInterfaces = getFcInternalInterfaces(itfs);
        }
        return singletonControlInterfaces;        
    }

    private List collectionInterfaces;
    
    /**
     * @return  the list of internal collection Interface instances implemented
     * by this component
     */
    private List getFcInternalCollectionInterfaces() {
        if( collectionInterfaces == null ) {
            ComponentItf c =
                (ComponentItf) FeatureHelper.getFcImplementation(compctrl);
            List itfs = c.getFcCollectionInterfaces();
            collectionInterfaces = getFcInternalInterfaces(itfs);
        }
        return collectionInterfaces;        
    }

    /**
     * Given a list of external Interface instances,
     * return the list of internal ones.
     */
    private static List getFcInternalInterfaces( List itfs ) {
        
        List externalItfs = new ArrayList(itfs.size());
        for (Iterator iter = itfs.iterator(); iter.hasNext();) {
            
            // Clone business Interfaces
            Interface itf = (Interface) iter.next();
            String itfName = itf.getFcItfName();
            if( ! FractalHelper.isFcControllerInterfaceName(itfName) ) {
                
                InterfaceImpl clone =
                    (InterfaceImpl) ((CloneableItf)itf).clone();
                
                // The clone Interface is internal
                clone.setFcInternalItf(true);
                
                // Change the type of the clone Interface (client <-> server)
                InterfaceType it = (InterfaceType) itf.getFcItfType();
                InterfaceType cloneit = 
                    new InterfaceTypeImpl(
                            it.getFcItfName(),
                            it.getFcItfSignature(),
                            ! it.isFcClientItf(), // client <-> server
                            it.isFcOptionalItf(),
                            it.isFcCollectionItf() );
                clone.setFcItfType(cloneit);
                
                externalItfs.add(clone);                
            }
        }
        
        return externalItfs;
    }
    
    /**
     * Return true if the given src component is contained in the given dst
     * component.
     * 
     * @param recurse  true if the search must be recursive
     * @param src      the source component
     * @param dst      the destination component
     * @return         true if src is contained in dst
     */
    private static boolean isFcContainedIn(
            boolean recurse, Component src, Component dst ) {
        
        ContentController cc = null;
        try {
            cc = org.objectweb.fractal.util.Fractal.getContentController(dst);
        }
        catch(NoSuchInterfaceException nsie) {
            // dst is not a composite
            // Should we propagate NoSuchInterfaceException instead ?
            return false;
        }
        
        Component[] contained = cc.getFcSubComponents();
        for (int i = 0; i < contained.length; i++) {
            if( contained[i].equals(src) ) {
                return true;
            }
            // Recursive search if needed
            if( recurse ) {
                if( isFcContainedIn(recurse,src,contained[i]) ) {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    
    // --------------------------------------------------------------
    // Controller implementation
    // --------------------------------------------------------------
    
    private Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        // Indeed nothing
    }

    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {
        
        // Get the content of the composite.
        // Recursively instantiate contained component and add them to dst.
        ContentController newCc = FractalHelper.getContentController(dst);
        
        // Map storing associations between components in contained in the
        // current composite and their associated newly created component.
        Map newContent = (Map) hints;
        
        Component[] subs = getFcSubComponents();
        
        for (int i = 0; i < subs.length; i++) {
            Component sub = (Component) newContent.get(subs[i]);
            if( sub == null ) {
                throw new CloneCtrlException(
                        "Contained component "+subs[i]+
                        " is associated to a null instance");
            }
            try {
                newCc.addFcSubComponent(sub);
            }
            catch( Exception e ) {
                throw new CloneCtrlException(e);
            }
        }
    }
}
