/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.control.factory;

import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;

import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.control.attribute.AttributeControllerItf;
import aokell.lib.control.component.ComponentDescItf;
import aokell.lib.util.FractalHelper;

/**
 * Root implementation for primitive and composite factories.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public abstract class AbstractFactoryImpl
    implements FactoryItf, Controller {
    
    public AbstractFactoryImpl() {
    }
    
    
    // --------------------------------------------------------------
    // Factory interface
    // --------------------------------------------------------------
    
    public Object getFcControllerDesc() {
        ComponentDescItf cdi =
            (ComponentDescItf) FractalHelper.getFcInterface(compctrl,"/desc");
        return cdi.getControllerDesc();
    }
    
    public Object getFcContentDesc() {
        ComponentDescItf cdi =
            (ComponentDescItf) FractalHelper.getFcInterface(compctrl,"/desc");
        return cdi.getContentDesc();
    }
    
    public Type getFcInstanceType() {
        return compctrl.getFcType();
    }
    
    abstract public Component newFcInstance() throws InstantiationException;
    
    
    // --------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------
    
    /**
     * Copy the state of the attribute controller which is in the same membrane
     * as the current factory controller into the attribute controller of the
     * newly created component.
     * 
     * @param newComponent  the newly created component
     * @param newContent    all newly created components
     */
    protected void copyAttributeControllerState(
            Component component, Component newComponent, Map newContent )
    throws CloneCtrlException {
        
        Controller ctrl = null;
        
        try {
            ctrl = (Controller)
                component.getFcInterface(AttributeControllerItf.NAME);
        }
        catch( Exception e ) {
            // The component is not parametric
            return;
        }
        
        ctrl.cloneFcCtrl(newComponent,newContent);
    }
    
    
    // --------------------------------------------------------------
    // Controller implementation
    // --------------------------------------------------------------
    
    protected Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        // Indeed nothing
    }

    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {
        // Indeed nothing
    }
}
