/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 * Contributors: Romain Rouvoy
 * -----------------------------------------------------------------------------
 * $Id: AbstractLifeCycleControllerImpl.java,v 1.12 2005/12/31 15:48:34 seinturier Exp $
 * -----------------------------------------------------------------------------
 */

package aokell.lib.control.lifecycle;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;

import aokell.lib.control.CloneCtrlException;
import aokell.lib.control.Controller;
import aokell.lib.util.FeatureHelper;
import aokell.lib.util.FractalHelper;

/**
 * Root class for life cyle controller implementations.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @author <a href="mailto:Romain.Rouvoy@lifl.fr">Romain Rouvoy</a>
 */
public abstract class AbstractLifeCycleControllerImpl
    implements LifeCycleControllerItf, Controller {
    
    protected boolean started = false;

    // --------------------------------------------------------------
    // LifeCycleController interface
    // --------------------------------------------------------------
    
    /* (non-Javadoc)
     * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
     */
    public String getFcState() {
        return (started ? STARTED : STOPPED);
    }

    /* (non-Javadoc)
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    public void startFc() throws IllegalLifeCycleException {
        
        Component compctrl = FeatureHelper.getFcImplementation(this.compctrl);
        
        ComponentType ct = (ComponentType) compctrl.getFcType();
        InterfaceType[] its = ct.getFcInterfaceTypes();
        BindingController bc = FractalHelper.getBindingController(compctrl);
        
        // Mandatory interfaces must be bound before starting the component
        for (int i = 0; i < its.length; i++) {
            if( its[i].isFcClientItf() && ! its[i].isFcOptionalItf() &&
                // Collection interfaces are excluded from the test
                // See TestLifeCycleController.testRecursiveStartStop()
                ! its[i].isFcCollectionItf()) {
                String itfName = its[i].getFcItfName();
                Object bound = null;
                try {
                    bound = bc.lookupFc(itfName);                    
                } catch( NoSuchInterfaceException nsie ) {
                    throw new IllegalLifeCycleException("No such interface "+itfName);
                }
                if( bound == null ) {
                    throw new IllegalLifeCycleException(
                            "Mandatory interface "+itfName+
                            " must be bound before starting component "+
                            FractalHelper.getContent(compctrl));
                }
                // TODO mandatory server interfaces should be tested too
                // Client interfaces bound to the server interfaces should be recorded
            }
        }
        if (!setFcStarted())
            throw new IllegalLifeCycleException("The component is already "+getFcState()+".");
    }
    
    /* (non-Javadoc)
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    public void stopFc() throws IllegalLifeCycleException {        
        if (!setFcStopped())
            throw new IllegalLifeCycleException("The component is already "+getFcState()+".");
    }


    // --------------------------------------------------------------
    // LifeCycleControllerItf interface
    // --------------------------------------------------------------
    
    /* (non-Javadoc)
     * @see aokell.lib.control.lifecycle.LifeCycleControllerItf#isFcStarted()
     */
    public boolean isFcStarted() {
        return started;
    }
    
    /* (non-Javadoc)
     * @see aokell.lib.control.lifecycle.LifeCycleControllerItf#setFcStarted()
     */
    public boolean setFcStarted()throws IllegalLifeCycleException {
        if (!started) {
            started = true;
            return true;
        } 
        return false;
    }
    
    /* (non-Javadoc)
     * @see aokell.lib.control.lifecycle.LifeCycleControllerItf#setFcStopped()
     */
    public boolean setFcStopped()throws IllegalLifeCycleException {
        if (started) {
            started = false;
            return true;
        } 
        return false;
    }
    
    
    // --------------------------------------------------------------
    // Controller implementation
    // --------------------------------------------------------------
    
    protected Component compctrl;
    
    /**
     * Set the reference towards the component controller associated to this
     * controller.
     */
    public void setFcCompCtrl( Component compctrl ) {
        this.compctrl = compctrl;
    }
    
    /**
     * Initialize the controller.
     */
    public void initFcCtrl() {
        
        /*
         * Optimization: direct reference the lifecycle controller.
         */
        Object content = FractalHelper.getContent(compctrl);
        
        if( content instanceof LifeCycleControllerSetterGetterItf ) {
            ((LifeCycleControllerSetterGetterItf) content).
                setFcLifeCycleController(this);
        }
    }
    
    /**
     * Clone the controller state from the current component to another one.
     * This method may receive some hints on how to do this, or provide some
     * hints on how this has been done. For instance, the hints may be a map
     * that is read and/or written by the controller. The raison d'etre of
     * these hints is that when its state is cloned, a controller may produce
     * results that are needed by other controllers. For instance the content
     * controller creates new instances, and put in the hints a map associating
     * old instances and their newly created conterparts.
     * 
     * @param dst    the destination component
     * @param hints  hints for performing the operation
     */
    public void cloneFcCtrl( Component dst, Object hints )
    throws CloneCtrlException {

        // Start dst if the current component is started.
        // If the current component is stopped, do nothing (dst is supposed to
        // be in its default state which is stopped). 
        if( started ) {
            LifeCycleController lc = FractalHelper.getLifeCycleController(dst);
            try {
                lc.startFc();                    
            } catch( IllegalLifeCycleException ilce ) {
                /*
                 * When a cloned composite is started, all its cloned sub
                 * components are also started. Hence, when the cloneCtrl
                 * method is executed for these last components, an
                 * IllegalLifeCycleException is thrown (they are already
                 * started). We catch it here and do nothing as this is a
                 * normal behavior.
                 */
            }
        }
    }
}