/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.lib.util;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.control.SuperController;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.type.TypeFactory;

import aokell.lib.control.binding.BindingControllerItf;
import aokell.lib.control.factory.FactoryItf;
import aokell.lib.control.superc.SuperControllerItf;

/**
 * Helper class that wraps calls to o.o.f.a.Fractal and that throws runtime
 * exception instead of checked exceptions.
 * Various other utility methods.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class FractalHelper {

    public static Object getFcInterface( Component c, String itfName ) {
        try {
            return c.getFcInterface(itfName);
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie);
        }        
    }
    
    /**
     * Returns the {@link AttributeController} interface of the given component.
     *
     * @param c  a component.
     * @return the {@link AttributeController} interface ofthe given component.
     */
    public static AttributeController getAttributeController(Component c) {
        return (AttributeController) getFcInterface(c,"attribute-controller");
    }
    
    /**
     * Returns the {@link BindingControllerItf}interface of the given component.
     * 
     * @param c  a component.
     * @return the {@link BindingControllerItf} interface of the given component.
     */
    public static BindingControllerItf getBindingController(Component c) {
        return (BindingControllerItf) getFcInterface(c,"binding-controller");
    }

    /**
     * Returns the {@link ContentController} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link ContentController} interface of the given component.
     */
    public static ContentController getContentController(Component c) {
        return (ContentController) getFcInterface(c,"content-controller");
    }

    /**
     * Returns the {@link SuperControllerItf} interface of the given component.
     * 
     * @param c  a component.
     * @return the {@link SuperControllerItf} interface of the given component.
     */
    public static SuperControllerItf getSuperController(Component c) {
        return (SuperControllerItf) getFcInterface(c,"super-controller");
    }

    /**
     * Returns the {@link NameController} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link NameController} interface of the given component.
     */
    public static NameController getNameController(Component c) {
        return (NameController) getFcInterface(c,"name-controller");
    }

    /**
     * Returns the {@link LifeCycleController} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link LifeCycleController} interface of the given component.
     * @throws NoSuchInterfaceException
     *             if there is no such interface.
     */
    public static LifeCycleController getLifeCycleController(Component c) {
        return (LifeCycleController) getFcInterface(c,"lifecycle-controller");
    }

    /**
     * Returns the {@link FactoryItf} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link FactoryItf} interface of the given component.
     */
    public static FactoryItf getFactory(Component c) {
        return (FactoryItf) getFcInterface(c,"factory");
    }

    /**
     * Returns the {@link GenericFactory} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link GenericFactory} interface of the given component.
     */
    public static GenericFactory getGenericFactory(Component c) {
        return (GenericFactory) getFcInterface(c,"generic-factory");
    }

    /**
     * Returns the {@link TypeFactory} interface of the given component.
     * 
     * @param c a component.
     * @return the {@link TypeFactory} interface of the given component.
     */
    public static TypeFactory getTypeFactory(Component c) {
        return (TypeFactory) getFcInterface(c,"type-factory");
    }
    
    public static Object getContent(Component c) {
        return getFcInterface(c,"/content");
    }
    
    /**
     * Bind the given destination interface to the given client interface name.
     * 
     * @param bc       the binding controller
     * @param itfName  the client interface name
     * @param dstItf   the destination interface
     */
    public static void bindFc(
            BindingController bc, String itfName, Object dstItf ) {
        try {
            bc.bindFc(itfName,dstItf);
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie);
        }
        catch( IllegalBindingException ibe ) {
            throw new RuntimeException(ibe);
        }
        catch( IllegalLifeCycleException ilce) {
            throw new RuntimeException(ilce);
        }
    }
    
    public static Object lookupFc( BindingController bc, String itfName ) {
        try {
            return bc.lookupFc(itfName);
        }
        catch( NoSuchInterfaceException nsie ) {
            throw new RuntimeException(nsie);
        }
    }
    
    /**
     * Return true if the specified name corresponds to the name of a Fractal
     * controller interface name.
     * Note: factory is not considered as a controller interface but as a
     * business one.
     */
    public static boolean isFcControllerInterfaceName(String itfName) {
        if( itfName.equals("component") ||
            itfName.endsWith("-controller") ) {
            return true;
        }
        return false;
    }
    
    /**
     * Follow bindings and return the interface associated to the component
     * located at the end of the binding chain.
     */
    public static Interface followBindingsUpToLastInterface( Interface start ) {
        
        Object next = null;
        Component comp = null;
        
        do {
            comp = start.getFcItfOwner();            
            BindingController bc = getBindingController(comp);
            try {
                next = bc.lookupFc(start.getFcItfName());
                if( next != null ) {
                    start = (Interface) next;
                }
            }
            catch( NoSuchInterfaceException nsie ) {
                next = null;
            }
        }
        while( next != null );
        
        return start;
    }

    /**
     * Return the fully-qualified name of a component, i.e. a slash-separated
     * string of names reflecting the hierarchy of nested components.
     */
    public static String getFcFullyQualifiedName( Component comp ) {
        
        SuperController sc = getSuperController(comp);
        Component[] supers = sc.getFcSuperComponents();
        if( supers.length == 0 ) {
            return "/"+getNameController(comp).getFcName();
        }
        
        return
            getFcFullyQualifiedName(supers[0])+"/"+
            getNameController(comp).getFcName();
    }
}
