/***
 * AOKell
 * Copyright (C) 2005 INRIA, France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package aokell.tools.content;


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;

import aokell.AOKellMembranes;
import aokell.lib.Content;
import aokell.lib.membrane.MembraneDef;
import aokell.lib.membrane.Membranes;

/**
 * Starting from an ADL description, this tool determines component content
 * classes and generates the AspectJ aspect which introduces the
 * aokell.lib.Content marker interface into those classes.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 */
public class ContentResolver {
    
    /** The directory where the generated file is stored. */
    final private static String generatedDirname = "generated";
    
    private static void usage() {
        System.out.println(
                "java "+ContentResolver.class.getName()+" .fractal file[s]");
        System.out.println(
                "Description: parse ADL architectures and extract Java content classes");
    }
    
    public static void main(String[] args) throws Exception {
        
        if( args.length == 0 ) {
            usage();
            return;
        }
        
        Membranes.get().register(AOKellMembranes.MEMBRANES);
        new ContentResolver(args).compile();
    }
    
    public ContentResolver( String[] adls ) {
        this.adls = adls;
    }
    
    private String[] adls;
    
    
    public void compile() throws ADLException, IOException {        
        for (int i = 0; i < adls.length; i++) {
            int lastdot = adls[i].lastIndexOf('.');
            String packagename =
                (lastdot==-1) ? "" : adls[i].substring(0,lastdot);
            String classname = adls[i].substring(lastdot+1);
            resolve(packagename,classname);
        }        
    }
    
    protected void resolve( String packagename, String classname )
    throws ADLException, IOException {
                
        String dirname = generatedDirname+"/"+packagename.replace('.','/');
        String aspectname = "A"+classname;
        String filename = dirname+"/"+aspectname+".aj";
        new File(dirname).mkdirs();
        System.out.println("Writing into: "+filename);
        
        FileWriter fw = new FileWriter(filename);
        PrintWriter pw = new PrintWriter(fw);
        
        prologue(pw,packagename,aspectname);        
        
        // contents is of type Map<Object,List<Object>>
        // <content desc, List<controller desc>>
        Map contents = new HashMap();
        
        Map context = new HashMap();
        context.put("contents",contents);

        // Backend.fractal must be stored in the same package as ContentResolver 
        Factory f = FactoryFactory.getFactory(
                ContentResolver.class.getPackage().getName()+".Backend");
        f.newComponent(packagename+"."+classname,context);
        
        Set set = contents.entrySet();
        for (Iterator iter = set.iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            Object contentDesc = entry.getKey();
            List ctrlDescs = (List) entry.getValue();
            
            if( ctrlDescs.size() == 0 ) {
                throw new RuntimeException("This shouldn't happen!");
            }
            else if( ctrlDescs.size() == 1 ) {
                // Regular case
                Object ctrlDesc = ctrlDescs.get(0);
                generateClause(pw,contentDesc,ctrlDesc);
            }
            else {
                System.out.println(
                        "Error: "+contentDesc+
                        " is associated with more than 1 controller description");
            }
        }

        epilogue(pw);
        
        pw.close();
        fw.close();
        
        System.out.println("Done.");
    }
    
    protected void generateClause(
            PrintWriter pw, Object contentDesc, Object ctrlDesc )
    throws IOException {
        
        MembraneDef mdef = Membranes.get().getMembraneDef(ctrlDesc);
        if( mdef == null ) {
            throw new RuntimeException(
                    "Unknown controller description: "+ctrlDesc);
        }
        
        for (int i = 0; i < mdef.intercepts.length; i++) {
            Class intercept = mdef.intercepts[i];
            
            String msg = contentDesc+" implements "+intercept.getName();

            pw.print("  declare parents: ");
            pw.print(msg);
            pw.print(";  // ");
            pw.println(ctrlDesc);
            
            System.out.println("  "+msg);
            
            pw.print("  declare parents: ");
            pw.print(contentDesc);
            pw.println(" implements Content;");
        }
        
    }
    
    protected void prologue(
            PrintWriter pw, String packagename, String aspectname )
    throws IOException {
        
        pw.println("/*");
        pw.print(" * Generated by AOKell ContentResolver on: ");
        pw.println(new Date().toString());
        pw.println(" */");
        pw.println();
        
        pw.println("package "+packagename+";");
        pw.println();
        pw.println("import "+Content.class.getName());
        pw.println();
        pw.println("public aspect "+aspectname+" {");
    }

    protected void epilogue( PrintWriter pw ) throws IOException {
        pw.println("}");
    }
}
